import numpy as np
import copy
from scipy.interpolate import interp1d
from lenstronomy.LightModel.light_model import LightModel

__all__ = ["LightProfile"]


class LightProfile(object):
    """Class to deal with the light distribution for GalKin.

    In particular, this class allows for:
     - (faster) interpolated calculation for a given profile (for a range that the Jeans equation is computed)
     - drawing 3d and 2d distributions from a given (spherical) profile
       (within bounds where the Jeans equation is expected to be accurate)
     - 2d projected profiles within the 3d integration range (truncated)
    """

    def __init__(
        self,
        profile_list,
        interpol_grid_num=2000,
        max_interpolate=1000,
        min_interpolate=0.001,
        max_draw=None,
    ):
        """

        :param profile_list: list of light profiles for LightModel module (must support light_3d() functionalities)
        :param interpol_grid_num: int; number of interpolation steps (logarithmically between min and max value)
        :param max_interpolate: float; maximum interpolation of 3d light profile
        :param min_interpolate: float; minimum interpolate (and also drawing of light profile)
        :param max_draw: float; (optional) if set, draws up to this radius, else uses max_interpolate value
        """

        self.light_model = LightModel(light_model_list=profile_list)
        self._interp_grid_num = interpol_grid_num
        self._max_interpolate = max_interpolate
        self._min_interpolate = min_interpolate
        if max_draw is None:
            max_draw = max_interpolate
        self._max_draw = max_draw

    def light_3d(self, r, kwargs_list):
        """Three-dimensional light profile.

        :param r: 3d radius
        :param kwargs_list: list of keyword arguments of light profiles (see
            LightModule)
        :return: flux per 3d volume at radius r
        """
        light_3d = self.light_model.light_3d(r, kwargs_list)
        return light_3d

    def light_3d_interp(self, r, kwargs_list, new_compute=False):
        """Interpolated three-dimensional light profile within bounds [min_interpolate,
        max_interpolate] in logarithmic units with interpol_grid_num numbers of
        interpolation steps.

        :param r: 3d radius
        :param kwargs_list: list of keyword arguments of light profiles (see
            LightModule)
        :param new_compute: boolean, if True, re-computes the interpolation (becomes
            valid with updated kwargs_list argument)
        :return: flux per 3d volume at radius r
        """
        if not hasattr(self, "_f_light_3d") or new_compute is True:
            r_array = np.logspace(
                np.log10(self._min_interpolate),
                np.log10(self._max_interpolate),
                self._interp_grid_num,
            )
            light_3d_array = self.light_model.light_3d(r_array, kwargs_list)
            light_3d_array[light_3d_array < 10 ** (-1000)] = 10 ** (-1000)
            f = interp1d(
                np.log(r_array),
                np.log(light_3d_array),
                fill_value=(np.log(light_3d_array[0]), -1000),
                bounds_error=False,
            )  # "extrapolate"
            self._f_light_3d = f
        return np.exp(self._f_light_3d(np.log(r)))

    def light_2d(self, R, kwargs_list):
        """Projected light profile (integrated to infinity in the projected axis)

        :param R: projected 2d radius
        :param kwargs_list: list of keyword arguments of light profiles (see
            LightModule)
        :return: projected surface brightness
        """
        kwargs_light_circularized = self._circularize_kwargs(kwargs_list)
        return self.light_model.surface_brightness(R, 0, kwargs_light_circularized)

    def _circularize_kwargs(self, kwargs_list):
        """

        :param kwargs_list: list of keyword arguments of light profiles (see LightModule)
        :return: circularized arguments
        """
        # TODO make sure averaging is done azimuthally
        if not hasattr(self, "_kwargs_light_circularized"):
            kwargs_list_copy = copy.deepcopy(kwargs_list)
            kwargs_list_new = []
            for kwargs in kwargs_list_copy:
                if "e1" in kwargs:
                    kwargs["e1"] = 0
                if "e2" in kwargs:
                    kwargs["e2"] = 0
                kwargs_list_new.append(
                    {
                        k: v
                        for k, v in kwargs.items()
                        if k not in ["center_x", "center_y"]
                    }
                )
            self._kwargs_light_circularized = kwargs_list_new
        return self._kwargs_light_circularized

    def _light_2d_finite_single(self, R, kwargs_list):
        """Projected light profile (integrated to FINITE 3d boundaries from the
        max_interpolate) for a single float number of R.

        :param R: projected 2d radius (between min_interpolate and max_interpolate)
        :param kwargs_list: list of keyword arguments of light profiles (see
            LightModule)
        :return: projected surface brightness
        """

        # here we perform a logarithmic integral
        stop = np.log10(
            np.maximum(
                np.sqrt(self._max_interpolate**2 - R**2),
                self._min_interpolate + 0.00001,
            )
        )
        x = np.logspace(
            start=np.log10(self._min_interpolate), stop=stop, num=self._interp_grid_num
        )
        r_array = np.sqrt(x**2 + R**2)
        flux_r = self.light_3d(r_array, kwargs_list)
        dlog_r = (np.log10(x[2]) - np.log10(x[1])) * np.log(10)
        flux_r *= dlog_r * x

        # linear integral
        # x = np.linspace(start=self._min_interpolate, stop=np.sqrt(self._max_interpolate ** 2 - R ** 2),
        #                num=self._interp_grid_num)
        # r_array = np.sqrt(x ** 2 + R ** 2)
        # dr = x[1] - x[0]
        # flux_r = self.light_3d(r_array + dr / 2, kwargs_circ)
        # dr = x[1] - x[0]
        # flux_r *= dr

        flux_R = np.sum(flux_r)
        # perform finite integral

        # out = integrate.quad(lambda x: self.light_3d(np.sqrt(R ** 2 + x ** 2), kwargs_circ), self._min_interpolate,
        #                     np.sqrt(self._max_interpolate**2 - R**2))
        # print(out_1, out, 'test')
        # flux_R = out[0]
        return flux_R * 2  # integral in both directions

    def light_2d_finite(self, R, kwargs_list):
        """Projected light profile (integrated to FINITE 3d boundaries from the
        max_interpolate)

        :param R: projected 2d radius (between min_interpolate and max_interpolate
        :param kwargs_list: list of keyword arguments of light profiles (see
            LightModule)
        :return: projected surface brightness
        """

        kwargs_circ = self._circularize_kwargs(kwargs_list)
        n = len(np.atleast_1d(R))
        if n <= 1:
            return self._light_2d_finite_single(R, kwargs_circ)
        else:
            light_2d = np.zeros(n)
            for i, R_i in enumerate(R):
                light_2d[i] = self._light_2d_finite_single(R_i, kwargs_circ)
            return light_2d

    def draw_light_2d_linear(self, kwargs_list, n=1, new_compute=False):
        """Constructs the CDF and draws from it random realizations of projected radii R
        The interpolation of the CDF is done in linear projected radius space.

        :param kwargs_list: list of keyword arguments of light profiles (see
            LightModule)
        :param n: int; number of draws
        :param new_compute: boolean, if True, re-computes the interpolation (becomes
            valid with updated kwargs_list argument)
        :return: draw of projected radius for the given light profile distribution
        """
        if not hasattr(self, "_light_cdf") or new_compute is True:
            r_array = np.linspace(
                self._min_interpolate, self._max_draw, self._interp_grid_num
            )
            cum_sum = np.zeros_like(r_array)
            sum_light = 0
            for i, r in enumerate(r_array):
                if i == 0:
                    cum_sum[i] = 0
                else:
                    sum_light += self.light_2d(r, kwargs_list) * r
                    cum_sum[i] = copy.deepcopy(sum_light)
            cum_sum_norm = cum_sum / cum_sum[-1]
            f = interp1d(cum_sum_norm, r_array)
            self._light_cdf = f
        cdf_draw = np.random.uniform(0.0, 1, n)
        r_draw = self._light_cdf(cdf_draw)
        return r_draw

    def draw_light_2d(self, kwargs_list, n=1, new_compute=False):
        """Constructs the CDF and draws from it random realizations of projected radii R
        CDF is constructed in logarithmic projected radius spacing.

        :param kwargs_list: light model keyword argument list
        :param n: int, number of draws per functino call
        :param new_compute: re-computes the interpolated CDF
        :return: realization of projected radius following the distribution of the light
            model
        """
        if not hasattr(self, "_light_cdf_log") or new_compute is True:
            r_array = np.logspace(
                np.log10(self._min_interpolate),
                np.log10(self._max_draw),
                self._interp_grid_num,
            )
            cum_sum = np.zeros_like(r_array)
            sum_light = 0
            for i, r in enumerate(r_array):
                if i == 0:
                    cum_sum[i] = 0
                else:
                    sum_light += self.light_2d(r, kwargs_list) * r * r
                    cum_sum[i] = copy.deepcopy(sum_light)
            cum_sum_norm = cum_sum / cum_sum[-1]
            f = interp1d(cum_sum_norm, np.log(r_array))
            self._light_cdf_log = f
        cdf_draw = np.random.uniform(0.0, 1, n)
        r_log_draw = self._light_cdf_log(cdf_draw)
        return np.exp(r_log_draw)

    def draw_light_3d(self, kwargs_list, n=1, new_compute=False):
        """Constructs the CDF and draws from it random realizations of 3D radii r.

        :param kwargs_list: light model keyword argument list
        :param n: int, number of draws per function call
        :param new_compute: re-computes the interpolated CDF
        :return: realization of projected radius following the distribution of the light
            model
        """
        if not hasattr(self, "_light_3d_cdf_log") or new_compute is True:
            r_array = np.logspace(
                np.log10(self._min_interpolate),
                np.log10(self._max_draw),
                self._interp_grid_num,
            )
            dlog_r = np.log10(r_array[1]) - np.log10(r_array[0])
            r_array_int = np.logspace(
                np.log10(self._min_interpolate) + dlog_r / 2,
                np.log10(self._max_draw) + dlog_r / 2,
                self._interp_grid_num,
            )
            cum_sum = np.zeros_like(r_array)
            sum_light = 0
            for i, r in enumerate(r_array_int[:-1]):
                # if i == 0:
                #    cum_sum[i] = 0
                # else:
                sum_light += (
                    self.light_3d(r, kwargs_list)
                    * r**2
                    * (r_array[i + 1] - r_array[i])
                )  # * r
                cum_sum[i + 1] = copy.deepcopy(sum_light)
            cum_sum_norm = cum_sum / cum_sum[-1]
            f = interp1d(cum_sum_norm, np.log(r_array))
            self._light_3d_cdf_log = f
        cdf_draw = np.random.uniform(0.0, 1, n)
        r_log_draw = self._light_3d_cdf_log(cdf_draw)
        return np.exp(r_log_draw)

    def delete_cache(self):
        """Deletes cached interpolation function of the CDF for a specific light
        profile.

        :return: None
        """
        if hasattr(self, "_light_cdf_log"):
            del self._light_cdf_log
        if hasattr(self, "_light_cdf"):
            del self._light_cdf
        if hasattr(self, "_f_light_3d"):
            del self._f_light_3d
        if hasattr(self, "_kwargs_light_circularized"):
            del self._kwargs_light_circularized



__author__ = "sibirrer"

import numpy as np
from scipy.interpolate import interp1d
import lenstronomy.GalKin.velocity_util as vel_util
from lenstronomy.GalKin.cosmo import Cosmo
from lenstronomy.GalKin.anisotropy import Anisotropy
from lenstronomy.LensModel.Profiles.spp import SPP
import lenstronomy.Util.constants as const
import math

__all__ = ["AnalyticKinematics"]


class AnalyticKinematics(Anisotropy):
    """Class to compute eqn 20 in Suyu+2010 with a Monte-Carlo from rendering from the
    light profile distribution and displacing them with a Gaussian seeing convolution.

    This class assumes spherical symmetry in light and mass distribution and
        - a Hernquist light profile (parameterised by the half-light radius)
        - a power-law mass profile (parameterized by the Einstein radius and logarithmic slop)

    The analytic equations for the kinematics in this approximation are presented e.g. in Suyu et al. 2010 and
    the spectral rendering approach to compute the seeing convolved slit measurement is presented in Birrer et al. 2016.
    The stellar anisotropy is parameterised based on Osipkov 1979; Merritt 1985.

    WARNING!!! Only supports Osipkov-Merritt anisotropy for now!

    All units are meant to be in angular arc seconds. The physical units are fold in through the angular diameter
    distances
    """

    def __init__(
        self,
        kwargs_cosmo,
        interpol_grid_num=100,
        log_integration=False,
        max_integrate=100,
        min_integrate=0.001,
    ):
        """

        :param kwargs_cosmo: keyword argument with angular diameter distances entering the Galkin.cosmo class
        :param interpol_grid_num: number of interpolations in radius to compute radial velocity dispersion
        :param log_integration: perform numerical integration in logarithmic space
        :param max_integrate: maximum radius of integration (in projected arc seconds)
        :param min_integrate: minimum drawing/calculation of velocity dispersion (in projected arc seconds)
        """

        self._interp_grid_num = interpol_grid_num
        self._log_int = log_integration
        self._max_integrate = (
            max_integrate  # maximal integration (and interpolation) in units of arcsecs
        )
        self._min_integrate = (
            min_integrate  # min integration (and interpolation) in units of arcsecs
        )
        self._max_interpolate = max_integrate  # we chose to set the interpolation range to the integration range
        self._min_interpolate = min_integrate  # we chose to set the interpolation range to the integration range

        self._cosmo = Cosmo(**kwargs_cosmo)
        self._spp = SPP()
        Anisotropy.__init__(self, anisotropy_type="OM")

    def _rho0_r0_gamma(self, theta_E, gamma):
        # equation (14) in Suyu+ 2010
        return (
            -1
            * math.gamma(gamma / 2)
            / (np.sqrt(np.pi) * math.gamma((gamma - 3) / 2.0))
            * theta_E**gamma
            / self._cosmo.arcsec2phys_lens(theta_E)
            * self._cosmo.epsilon_crit
            * const.M_sun
            / const.Mpc**3
        )

    @staticmethod
    def draw_light(kwargs_light):
        """

        :param kwargs_light: keyword argument (list) of the light model
        :return: 3d radius (if possible), 2d projected radius, x-projected coordinate, y-projected coordinate
        """
        if "a" not in kwargs_light:
            kwargs_light["a"] = 0.551 * kwargs_light["r_eff"]
        a = kwargs_light["a"]
        r = vel_util.draw_hernquist(a)
        R, x, y = vel_util.project2d_random(r)
        return r, R, x, y

    def _sigma_s2(self, r, R, r_ani, a, gamma, rho0_r0_gamma):
        """Projected velocity dispersion :param r: 3d radius of the light tracer
        particle :param R: 2d projected radius of the light tracer particle :param
        r_ani: anisotropy radius :param a: scale of the Hernquist light profile :param
        gamma: power-law slope of the mass profile :param rho0_r0_gamma: combination of
        Einstein radius and power-law slope as equation (14) in Suyu+ 2010 :return:
        projected velocity dispersion."""
        beta = self.beta_r(r, **{"r_ani": r_ani})
        return (1 - beta * R**2 / r**2) * self._sigma_r2_interp(
            r, a, gamma, rho0_r0_gamma, r_ani
        )

    def sigma_s2(self, r, R, kwargs_mass, kwargs_light, kwargs_anisotropy):
        """Returns unweighted los velocity dispersion for a specified projected radius,
        with weight 1.

        :param r: 3d radius (not needed for this calculation)
        :param R: 2d projected radius (in angular units of arcsec)
        :param kwargs_mass: mass model parameters (following lenstronomy lens model
            conventions)
        :param kwargs_light: deflector light parameters (following lenstronomy light
            model conventions)
        :param kwargs_anisotropy: anisotropy parameters, may vary according to
            anisotropy type chosen. We refer to the Anisotropy() class for details on
            the parameters.
        :return: line-of-sight projected velocity dispersion at projected radius R from
            3d radius r
        """
        a, gamma, rho0_r0_gamma, r_ani = self._read_out_params(
            kwargs_mass, kwargs_light, kwargs_anisotropy
        )
        return self._sigma_s2(r, R, r_ani, a, gamma, rho0_r0_gamma), 1

    def sigma_r2(self, r, kwargs_mass, kwargs_light, kwargs_anisotropy):
        """Equation (19) in Suyu+ 2010.

        :param r: 3d radius
        :param kwargs_mass: mass profile keyword arguments
        :param kwargs_light: light profile keyword arguments
        :param kwargs_anisotropy: anisotropy keyword arguments
        :return: velocity dispersion in [m/s]
        """
        a, gamma, rho0_r0_gamma, r_ani = self._read_out_params(
            kwargs_mass, kwargs_light, kwargs_anisotropy
        )
        return self._sigma_r2(r, a, gamma, rho0_r0_gamma, r_ani)

    def _read_out_params(self, kwargs_mass, kwargs_light, kwargs_anisotropy):
        """Reads the relevant parameters out of the keyword arguments and transforms
        them to the conventions used in this class.

        :param kwargs_mass: mass profile keyword arguments
        :param kwargs_light: light profile keyword arguments
        :param kwargs_anisotropy: anisotropy keyword arguments
        :return: a (Rs of Hernquist profile), gamma, rho0_r0_gamma, r_ani
        """
        if "a" not in kwargs_light:
            kwargs_light["a"] = 0.551 * kwargs_light["r_eff"]
        if "rho0_r0_gamma" not in kwargs_mass:
            kwargs_mass["rho0_r0_gamma"] = self._rho0_r0_gamma(
                kwargs_mass["theta_E"], kwargs_mass["gamma"]
            )
        a = kwargs_light["a"]
        gamma = kwargs_mass["gamma"]
        rho0_r0_gamma = kwargs_mass["rho0_r0_gamma"]
        r_ani = kwargs_anisotropy["r_ani"]
        return a, gamma, rho0_r0_gamma, r_ani

    def _sigma_r2(self, r, a, gamma, rho0_r0_gamma, r_ani):
        """Equation (19) in Suyu+ 2010."""
        # first term
        prefac1 = 4 * np.pi * const.G * a ** (-gamma) * rho0_r0_gamma / (3 - gamma)
        prefac2 = r * (r + a) ** 3 / (r**2 + r_ani**2)
        # TODO check whether interpolation functions can speed this up
        hyp1 = vel_util.hyp_2F1(a=2 + gamma, b=gamma, c=3 + gamma, z=1.0 / (1 + r / a))
        hyp2 = vel_util.hyp_2F1(a=3, b=gamma, c=1 + gamma, z=-a / r)
        fac = r_ani**2 / a**2 * hyp1 / (
            (2 + gamma) * (r / a + 1) ** (2 + gamma)
        ) + hyp2 / (gamma * (r / a) ** gamma)
        return (
            prefac1 * prefac2 * fac * (const.arcsec * self._cosmo.dd * const.Mpc) ** 2
        )

    def _sigma_r2_interp(self, r, a, gamma, rho0_r0_gamma, r_ani):
        """

        :param r:
        :param a:
        :param gamma:
        :param rho0_r0_gamma:
        :param r_ani:
        :return:
        """
        if not hasattr(self, "_interp_sigma_r2"):
            min_log = np.log10(self._min_integrate)
            max_log = np.log10(self._max_integrate)
            r_array = np.logspace(min_log, max_log, self._interp_grid_num)
            I_R_sigma2_array = []
            for r_i in r_array:
                I_R_sigma2_array.append(
                    self._sigma_r2(r_i, a, gamma, rho0_r0_gamma, r_ani)
                )
            self._interp_sigma_r2 = interp1d(
                np.log(r_array), np.array(I_R_sigma2_array), fill_value="extrapolate"
            )
        return self._interp_sigma_r2(np.log(r))

    def grav_potential(self, r, kwargs_mass):
        """Gravitational potential in SI units.

        :param r: radius (arc seconds)
        :param kwargs_mass:
        :return: gravitational potential
        """
        theta_E = kwargs_mass["theta_E"]
        gamma = kwargs_mass["gamma"]
        mass_dimless = self._spp.mass_3d_lens(r, theta_E, gamma)
        mass_dim = (
            mass_dimless
            * const.arcsec**2
            * self._cosmo.dd
            * self._cosmo.ds
            / self._cosmo.dds
            * const.Mpc
            * const.c**2
            / (4 * np.pi * const.G)
        )
        grav_pot = -const.G * mass_dim / (r * const.arcsec * self._cosmo.dd * const.Mpc)
        return grav_pot

    def delete_cache(self):
        """Deletes temporary cache tight to a specific model.

        :return:
        """
        if hasattr(self, "_interp_sigma_r2"):
            del self._interp_sigma_r2



import numpy as np
from lenstronomy.Util import util
from lenstronomy.LightModel.light_model import LightModel

__all__ = ["light2mass_interpol"]


def light2mass_interpol(
    lens_light_model_list,
    kwargs_lens_light,
    numPix=100,
    deltaPix=0.05,
    subgrid_res=5,
    center_x=0,
    center_y=0,
):
    """Takes a lens light model and turns it numerically in a lens model (with all
    lensmodel quantities computed on a grid). Then provides an interpolated grid for the
    quantities.

    :param kwargs_lens_light: lens light keyword argument list
    :param numPix: number of pixels per axis for the return interpolation
    :param deltaPix: interpolation/pixel size
    :param center_x: center of the grid
    :param center_y: center of the grid
    :param subgrid_res: subgrid for the numerical integrals
    :return: keyword arguments for 'INTERPOL' lens model
    """
    # make super-sampled grid
    x_grid_sub, y_grid_sub = util.make_grid(
        numPix=numPix * 5, deltapix=deltaPix, subgrid_res=subgrid_res
    )
    import lenstronomy.Util.mask_util as mask_util

    mask = mask_util.mask_azimuthal(x_grid_sub, y_grid_sub, center_x, center_y, r=1)
    x_grid, y_grid = util.make_grid(numPix=numPix, deltapix=deltaPix)
    # compute light on the subgrid
    lightModel = LightModel(light_model_list=lens_light_model_list)
    flux = lightModel.surface_brightness(x_grid_sub, y_grid_sub, kwargs_lens_light)
    flux_norm = np.sum(flux[mask == 1]) / np.sum(mask)
    flux /= flux_norm
    from lenstronomy.LensModel import convergence_integrals as integral

    # compute lensing quantities with subgrid
    convergence_sub = util.array2image(flux)
    f_x_sub, f_y_sub = integral.deflection_from_kappa_grid(
        convergence_sub, grid_spacing=deltaPix / float(subgrid_res)
    )
    f_sub = integral.potential_from_kappa_grid(
        convergence_sub, grid_spacing=deltaPix / float(subgrid_res)
    )
    # interpolation function on lensing quantities
    x_axes_sub, y_axes_sub = util.get_axes(x_grid_sub, y_grid_sub)
    from lenstronomy.LensModel.Profiles.interpol import Interpol

    interp_func = Interpol()
    interp_func.do_interp(x_axes_sub, y_axes_sub, f_sub, f_x_sub, f_y_sub)
    # compute lensing quantities on sparser grid
    x_axes, y_axes = util.get_axes(x_grid, y_grid)
    f_ = interp_func.function(x_grid, y_grid)
    f_x, f_y = interp_func.derivatives(x_grid, y_grid)
    # numerical differentials for second order differentials
    from lenstronomy.LensModel.lens_model import LensModel

    lens_model = LensModel(lens_model_list=["INTERPOL"])
    kwargs = [
        {
            "grid_interp_x": x_axes_sub,
            "grid_interp_y": y_axes_sub,
            "f_": f_sub,
            "f_x": f_x_sub,
            "f_y": f_y_sub,
        }
    ]
    f_xx, f_xy, f_yx, f_yy = lens_model.hessian(x_grid, y_grid, kwargs, diff=0.00001)
    kwargs_interpol = {
        "grid_interp_x": x_axes,
        "grid_interp_y": y_axes,
        "f_": util.array2image(f_),
        "f_x": util.array2image(f_x),
        "f_y": util.array2image(f_y),
        "f_xx": util.array2image(f_xx),
        "f_xy": util.array2image(f_xy),
        "f_yy": util.array2image(f_yy),
    }
    return kwargs_interpol



import numpy as np
from lenstronomy.LensModel.Profiles.gaussian_kappa import GaussianKappa
from lenstronomy.LensModel.Profiles.gaussian_ellipse_potential import (
    GaussianEllipsePotential,
)
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["MultiGaussianKappa", "MultiGaussianKappaEllipse"]


class MultiGaussianKappa(LensProfileBase):
    """"""

    param_names = ["amp", "sigma", "center_x", "center_y"]
    lower_limit_default = {"amp": 0, "sigma": 0, "center_x": -100, "center_y": -100}
    upper_limit_default = {"amp": 100, "sigma": 100, "center_x": 100, "center_y": 100}

    def __init__(self):
        self.gaussian_kappa = GaussianKappa()
        super(MultiGaussianKappa, self).__init__()

    def function(self, x, y, amp, sigma, center_x=0, center_y=0, scale_factor=1):
        """

        :param x:
        :param y:
        :param amp:
        :param sigma:
        :param center_x:
        :param center_y:
        :return:
        """
        f_ = np.zeros_like(x, dtype=float)
        for i in range(len(amp)):
            f_ += self.gaussian_kappa.function(
                x,
                y,
                amp=scale_factor * amp[i],
                sigma=sigma[i],
                center_x=center_x,
                center_y=center_y,
            )
        return f_

    def derivatives(self, x, y, amp, sigma, center_x=0, center_y=0, scale_factor=1):
        """

        :param x:
        :param y:
        :param amp:
        :param sigma:
        :param center_x:
        :param center_y:
        :return:
        """
        f_x, f_y = np.zeros_like(x, dtype=float), np.zeros_like(x, dtype=float)
        for i in range(len(amp)):
            f_x_i, f_y_i = self.gaussian_kappa.derivatives(
                x,
                y,
                amp=scale_factor * amp[i],
                sigma=sigma[i],
                center_x=center_x,
                center_y=center_y,
            )
            f_x += f_x_i
            f_y += f_y_i
        return f_x, f_y

    def hessian(self, x, y, amp, sigma, center_x=0, center_y=0, scale_factor=1):
        """

        :param x:
        :param y:
        :param amp:
        :param sigma:
        :param center_x:
        :param center_y:
        :return:
        """
        f_xx, f_yy, f_xy = (
            np.zeros_like(x, dtype=float),
            np.zeros_like(x, dtype=float),
            np.zeros_like(x, dtype=float),
        )
        for i in range(len(amp)):
            f_xx_i, f_xy_i, _, f_yy_i = self.gaussian_kappa.hessian(
                x,
                y,
                amp=scale_factor * amp[i],
                sigma=sigma[i],
                center_x=center_x,
                center_y=center_y,
            )
            f_xx += f_xx_i
            f_yy += f_yy_i
            f_xy += f_xy_i
        return f_xx, f_xy, f_xy, f_yy

    def density(self, r, amp, sigma, scale_factor=1):
        """

        :param r:
        :param amp:
        :param sigma:
        :return:
        """
        d_ = np.zeros_like(r, dtype=float)
        for i in range(len(amp)):
            d_ += self.gaussian_kappa.density(r, scale_factor * amp[i], sigma[i])
        return d_

    def density_2d(self, x, y, amp, sigma, center_x=0, center_y=0, scale_factor=1):
        """

        :param R:
        :param am:
        :param sigma_x:
        :param sigma_y:
        :return:
        """
        d_3d = np.zeros_like(x, dtype=float)
        for i in range(len(amp)):
            d_3d += self.gaussian_kappa.density_2d(
                x, y, scale_factor * amp[i], sigma[i], center_x, center_y
            )
        return d_3d

    def mass_3d_lens(self, R, amp, sigma, scale_factor=1):
        """

        :param R:
        :param amp:
        :param sigma:
        :return:
        """
        mass_3d = np.zeros_like(R, dtype=float)
        for i in range(len(amp)):
            mass_3d += self.gaussian_kappa.mass_3d_lens(
                R, scale_factor * amp[i], sigma[i]
            )
        return mass_3d


class MultiGaussianKappaEllipse(LensProfileBase):
    """"""

    param_names = ["amp", "sigma", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "amp": 0,
        "sigma": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 100,
        "sigma": 100,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self.gaussian_kappa = GaussianEllipsePotential()
        super(MultiGaussianKappaEllipse, self).__init__()

    def function(
        self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0, scale_factor=1
    ):
        """

        :param x:
        :param y:
        :param amp:
        :param sigma:
        :param center_x:
        :param center_y:
        :return:
        """
        f_ = np.zeros_like(x, dtype=float)
        for i in range(len(amp)):
            f_ += self.gaussian_kappa.function(
                x,
                y,
                amp=scale_factor * amp[i],
                sigma=sigma[i],
                e1=e1,
                e2=e2,
                center_x=center_x,
                center_y=center_y,
            )
        return f_

    def derivatives(
        self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0, scale_factor=1
    ):
        """

        :param x:
        :param y:
        :param amp:
        :param sigma:
        :param center_x:
        :param center_y:
        :return:
        """
        f_x, f_y = np.zeros_like(x, dtype=float), np.zeros_like(x, dtype=float)
        for i in range(len(amp)):
            f_x_i, f_y_i = self.gaussian_kappa.derivatives(
                x,
                y,
                amp=scale_factor * amp[i],
                sigma=sigma[i],
                e1=e1,
                e2=e2,
                center_x=center_x,
                center_y=center_y,
            )
            f_x += f_x_i
            f_y += f_y_i
        return f_x, f_y

    def hessian(self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0, scale_factor=1):
        """

        :param x:
        :param y:
        :param amp:
        :param sigma:
        :param center_x:
        :param center_y:
        :return:
        """
        f_xx, f_yy, f_xy = (
            np.zeros_like(x, dtype=float),
            np.zeros_like(x, dtype=float),
            np.zeros_like(x, dtype=float),
        )
        for i in range(len(amp)):
            f_xx_i, f_xy_i, _, f_yy_i = self.gaussian_kappa.hessian(
                x,
                y,
                amp=scale_factor * amp[i],
                sigma=sigma[i],
                e1=e1,
                e2=e2,
                center_x=center_x,
                center_y=center_y,
            )
            f_xx += f_xx_i
            f_yy += f_yy_i
            f_xy += f_xy_i
        return f_xx, f_xy, f_xy, f_yy

    def density(self, r, amp, sigma, e1, e2, scale_factor=1):
        """

        :param r:
        :param amp:
        :param sigma:
        :return:
        """
        d_ = np.zeros_like(r, dtype=float)
        for i in range(len(amp)):
            d_ += self.gaussian_kappa.density(
                r, scale_factor * amp[i], sigma[i], e1, e2
            )
        return d_

    def density_2d(
        self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0, scale_factor=1
    ):
        """

        :param R:
        :param am:
        :param sigma_x:
        :param sigma_y:
        :return:
        """
        d_3d = np.zeros_like(x, dtype=float)
        for i in range(len(amp)):
            d_3d += self.gaussian_kappa.density_2d(
                x, y, scale_factor * amp[i], sigma[i], e1, e2, center_x, center_y
            )
        return d_3d

    def mass_3d_lens(self, R, amp, sigma, e1, e2, scale_factor=1):
        """

        :param R:
        :param amp:
        :param sigma:
        :return:
        """
        mass_3d = np.zeros_like(R, dtype=float)
        for i in range(len(amp)):
            mass_3d += self.gaussian_kappa.mass_3d_lens(
                R, scale_factor * amp[i], sigma[i], e1, e2
            )
        return mass_3d



__author__ = "sibirrer"

import pytest
import numpy as np
import numpy.testing as npt

from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.Util import util


class TestNumerics(object):
    """Tests the source model routines."""

    def setup_method(self):
        self.lensModel = LensModel(["GAUSSIAN"])
        self.kwargs = [
            {
                "amp": 1.0 / 4.0,
                "sigma_x": 2.0,
                "sigma_y": 2.0,
                "center_x": 0.0,
                "center_y": 0.0,
            }
        ]

    def test_kappa(self):
        x, y = 1.0, 1.0
        output = self.lensModel.kappa(x, y, self.kwargs)
        output_num = self.lensModel.kappa(x, y, self.kwargs, diff=0.00001)
        npt.assert_almost_equal(output_num, output, decimal=5)

    def test_gamma(self):
        x, y = 1.0, 2.0
        output1, output2 = self.lensModel.gamma(x, y, self.kwargs)
        output1_num, output2_num = self.lensModel.gamma(x, y, self.kwargs, diff=0.00001)
        npt.assert_almost_equal(output1_num, output1, decimal=5)
        npt.assert_almost_equal(output2_num, output2, decimal=5)

    def test_magnification(self):
        x, y = 1.0, 1.0
        output = self.lensModel.magnification(x, y, self.kwargs)
        output_num = self.lensModel.magnification(x, y, self.kwargs, diff=0.00001)
        npt.assert_almost_equal(output_num, output, decimal=5)

    def test_differentials(self):
        x, y = util.make_grid(numPix=10, deltapix=0.5)
        f_xx, f_xy, f_yx, f_yy = self.lensModel.hessian(x, y, self.kwargs)
        f_xx_num, f_xy_num, f_yx_num, f_yy_num = self.lensModel.hessian(
            x, y, self.kwargs, diff=0.00001
        )

        npt.assert_almost_equal(f_xy_num, f_yx_num, decimal=5)
        npt.assert_almost_equal(f_xx_num, f_xx, decimal=5)
        npt.assert_almost_equal(f_xy_num, f_xy, decimal=5)
        npt.assert_almost_equal(f_yx_num, f_yx, decimal=5)
        npt.assert_almost_equal(f_yy_num, f_yy, decimal=5)

    def test_flexion(self):
        lensModel = LensModel(lens_model_list=["FLEXION"])
        g1, g2, g3, g4 = 0.01, 0.02, 0.03, 0.04
        kwargs = [{"g1": g1, "g2": g2, "g3": g3, "g4": g4}]
        f_xxx, f_xxy, f_xyy, f_yyy = lensModel.flexion(
            x=1.0, y=1.0, kwargs=kwargs, diff=0.0001
        )
        npt.assert_almost_equal(f_xxx, g1, decimal=2)
        npt.assert_almost_equal(f_xxy, g2, decimal=2)
        npt.assert_almost_equal(f_xyy, g3, decimal=2)
        npt.assert_almost_equal(f_yyy, g4, decimal=2)

        f_xxx, f_xxy, f_xyy, f_yyy = lensModel.flexion(
            x=1.0, y=1.0, kwargs=kwargs, diff=0.0001, hessian_diff=0.001
        )
        npt.assert_almost_equal(f_xxx, g1, decimal=2)
        npt.assert_almost_equal(f_xxy, g2, decimal=2)
        npt.assert_almost_equal(f_xyy, g3, decimal=2)
        npt.assert_almost_equal(f_yyy, g4, decimal=2)


class TestNumericsProfile(object):
    """Tests the second derivatives of various lens models."""

    def setup_method(self):
        pass

    def assert_differentials(self, lens_model, kwargs, potential=True):
        # lensModelNum = NumericLens(lens_model)
        diff = 0.000001
        # x, y = 1., 2.

        x = np.linspace(start=0.1, stop=5.5, num=10)
        y = np.zeros_like(x)

        lensModel = LensModel(lens_model)
        f_xx, f_xy, f_yx, f_yy = lensModel.hessian(x, y, [kwargs])
        f_xx_num, f_xy_num, f_yx_num, f_yy_num = lensModel.hessian(
            x, y, [kwargs], diff=diff
        )

        npt.assert_almost_equal(f_xx, f_xx_num, decimal=3)
        npt.assert_almost_equal(f_yy, f_yy_num, decimal=3)
        npt.assert_almost_equal(f_xy, f_xy_num, decimal=3)

        if potential is True:
            f_x, f_y = lensModel.alpha(x, y, [kwargs])
            f_x_num, f_y_num = lensModel.alpha(x, y, [kwargs], diff=diff)

            npt.assert_almost_equal(f_x, f_x_num, decimal=3)
            npt.assert_almost_equal(f_y, f_y_num, decimal=3)

        y = np.linspace(start=0.1, stop=5.5, num=10)
        x = np.zeros_like(y)

        lensModel = LensModel(lens_model)
        f_xx, f_xy, f_yx, f_yy = lensModel.hessian(x, y, [kwargs])
        f_xx_num, f_xy_num, f_yx_num, f_yy_num = lensModel.hessian(
            x, y, [kwargs], diff=diff
        )

        npt.assert_almost_equal(f_xx, f_xx_num, decimal=3)
        npt.assert_almost_equal(f_yy, f_yy_num, decimal=3)
        npt.assert_almost_equal(f_xy, f_xy_num, decimal=3)

        if potential is True:
            f_x, f_y = lensModel.alpha(x, y, [kwargs])
            f_x_num, f_y_num = lensModel.alpha(x, y, [kwargs], diff=diff)

            npt.assert_almost_equal(f_x, f_x_num, decimal=3)
            npt.assert_almost_equal(f_y, f_y_num, decimal=3)

    def test_gaussian(self):
        lens_model = ["GAUSSIAN"]
        kwargs = {
            "amp": 1.0 / 4.0,
            "sigma_x": 2.0,
            "sigma_y": 2.0,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        self.assert_differentials(lens_model, kwargs)
        kwargs = {
            "amp": 1.0 / 4.0,
            "sigma_x": 20.0,
            "sigma_y": 20.0,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        self.assert_differentials(lens_model, kwargs)
        kwargs = {
            "amp": 1.0 / 4.0,
            "sigma_x": 2.0,
            "sigma_y": 2.0,
            "center_x": 2.0,
            "center_y": 2.0,
        }
        self.assert_differentials(lens_model, kwargs)

    def test_gausian_kappa(self):
        kwargs = {"amp": 1.0 / 4.0, "sigma": 2.0, "center_x": 0.0, "center_y": 0.0}
        lens_model = ["GAUSSIAN_KAPPA"]
        self.assert_differentials(lens_model, kwargs)

    def test_gausian_ellipse_kappa(self):
        kwargs = {
            "amp": 1.0,
            "sigma": 1.0,
            "e1": 0.1,
            "e2": -0.1,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        lens_model = ["GAUSSIAN_ELLIPSE_KAPPA"]
        self.assert_differentials(lens_model, kwargs)

    def test_gausian_ellipse_potential(self):
        kwargs = {
            "amp": 1.0,
            "sigma": 2.0,
            "e1": 0.1,
            "e2": -0.1,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        lens_model = ["GAUSSIAN_ELLIPSE_POTENTIAL"]
        self.assert_differentials(lens_model, kwargs)

    def test_external_shear(self):
        kwargs = {"gamma1": 0.1, "gamma2": -0.1}
        lens_model = ["SHEAR"]
        self.assert_differentials(lens_model, kwargs)

    def test_reduce_shear(self):
        kwargs = {"gamma1": 0.1, "gamma2": -0.1}
        lens_model = ["SHEAR_REDUCED"]
        self.assert_differentials(lens_model, kwargs)

    def test_hessian(self):
        kwargs = {"f_xx": 0.1, "f_yy": -0.1, "f_xy": 0.1, "f_yx": 0.1}
        lens_model = ["HESSIAN"]
        self.assert_differentials(lens_model, kwargs)

    def test_mass_sheet(self):
        kwargs = {"kappa": 0.1}
        lens_model = ["CONVERGENCE"]
        self.assert_differentials(lens_model, kwargs)

    def test_sis(self):
        kwargs = {"theta_E": 0.5}
        lens_model = ["SIS"]
        self.assert_differentials(lens_model, kwargs)

    def test_flexion(self):
        kwargs = {"g1": 0.01, "g2": -0.01, "g3": 0.001, "g4": 0}
        lens_model = ["FLEXION"]
        self.assert_differentials(lens_model, kwargs)

    def test_nfw(self):
        kwargs = {"alpha_Rs": 0.1, "Rs": 5.0}
        lens_model = ["NFW"]
        self.assert_differentials(lens_model, kwargs)

    def test_tnfw(self):
        kwargs = {"alpha_Rs": 0.1, "Rs": 5.0, "r_trunc": 7}
        lens_model = ["TNFW"]
        self.assert_differentials(lens_model, kwargs)

        kwargs = {"Rs": 2.0, "alpha_Rs": 1.0, "r_trunc": 7}
        lens_model = ["TNFW"]
        self.assert_differentials(lens_model, kwargs)

    def test_tnfw_ellipse(self):
        lens_model = ["TNFW_ELLIPSE"]

        kwargs = {"alpha_Rs": 0.1, "Rs": 1.0, "r_trunc": 7, "e1": 0, "e2": 0}
        self.assert_differentials(lens_model, kwargs)

        kwargs = {"alpha_Rs": 0.1, "Rs": 1.0, "r_trunc": 7, "e1": 0.1, "e2": -0.02}
        self.assert_differentials(lens_model, kwargs)

        kwargs = {"Rs": 0.5, "alpha_Rs": 1.0, "r_trunc": 100, "e1": 0.1, "e2": -0.02}
        self.assert_differentials(lens_model, kwargs)

    def test_nfw_ellipse(self):
        kwargs = {"alpha_Rs": 0.1, "Rs": 5.0, "e1": 0.04, "e2": -0.04}
        lens_model = ["NFW_ELLIPSE"]
        self.assert_differentials(lens_model, kwargs)

    def test_nfw_ellipse_gauss_dec(self):
        kwargs = {"alpha_Rs": 0.1, "Rs": 5.0, "e1": 0.04, "e2": -0.04}
        lens_model = ["NFW_ELLIPSE_GAUSS_DEC"]
        self.assert_differentials(lens_model, kwargs)

    def test_ctnfw_gauss_dec(self):
        kwargs = {"rho_s": 5, "r_s": 5.0, "r_trunc": 10.0, "r_core": 0.3, "a": 2}
        lens_model = ["CTNFW_GAUSS_DEC"]
        self.assert_differentials(lens_model, kwargs)

    def test_point_mass(self):
        kwargs = {"theta_E": 2.0}
        lens_model = ["POINT_MASS"]
        self.assert_differentials(lens_model, kwargs)

    def test_sersic(self):
        kwargs = {"n_sersic": 0.5, "R_sersic": 1.5, "k_eff": 0.3}
        lens_model = ["SERSIC"]
        self.assert_differentials(lens_model, kwargs)

    def test_sersic_ellipse_gauss_dec(self):
        kwargs = {"n_sersic": 1.0, "R_sersic": 2.0, "k_eff": 1.0, "e1": 0.04, "e2": 0.0}
        lens_model = ["SERSIC_ELLIPSE_GAUSS_DEC"]
        self.assert_differentials(lens_model, kwargs)

    def test_sersic_ellipse_pot(self):
        kwargs = {
            "n_sersic": 2.0,
            "R_sersic": 0.5,
            "k_eff": 0.3,
            "e1": 0.04,
            "e2": -0.0,
        }
        lens_model = ["SERSIC_ELLIPSE_POTENTIAL"]
        self.assert_differentials(lens_model, kwargs)

    def test_shapelets_pot_2(self):
        kwargs = {"coeffs": [0, 1, 2, 3, 4, 5], "beta": 0.3}
        lens_model = ["SHAPELETS_CART"]
        self.assert_differentials(lens_model, kwargs)

    def test_sis_truncate(self):
        kwargs = {"theta_E": 0.5, "r_trunc": 2.0}
        lens_model = ["SIS_TRUNCATED"]
        self.assert_differentials(lens_model, kwargs)

    def test_spep(self):
        kwargs = {"theta_E": 0.5, "gamma": 1.9, "e1": 0.04, "e2": -0.1}
        lens_model = ["SPEP"]
        self.assert_differentials(lens_model, kwargs)

    def test_spp(self):
        kwargs = {"theta_E": 0.5, "gamma": 1.9}
        lens_model = ["SPP"]
        self.assert_differentials(lens_model, kwargs)

    def test_PJaffe(self):
        kwargs = {"sigma0": 1.0, "Ra": 0.2, "Rs": 2.0}
        lens_model = ["PJAFFE"]
        self.assert_differentials(lens_model, kwargs)

    def test_PJaffe_ellipse(self):
        kwargs = {"sigma0": 1.0, "Ra": 0.2, "Rs": 2.0, "e1": 0.04, "e2": -0.0}
        lens_model = ["PJAFFE_ELLIPSE"]
        self.assert_differentials(lens_model, kwargs)

    def test_Hernquist(self):
        kwargs = {"sigma0": 1.0, "Rs": 1.5}
        lens_model = ["HERNQUIST"]
        self.assert_differentials(lens_model, kwargs)

    def test_Hernquist_ellipse(self):
        kwargs = {"sigma0": 1.0, "Rs": 1.5, "e1": 0.04, "e2": -0.0}
        lens_model = ["HERNQUIST_ELLIPSE"]
        self.assert_differentials(lens_model, kwargs)

    def test_NIE(self):
        kwargs = {"theta_E": 2.0, "e1": 0.1, "e2": 0.0, "s_scale": 0.04}
        lens_model = ["NIE"]
        self.assert_differentials(lens_model, kwargs)

    def test_NIE_simple(self):
        kwargs = {"b": 2.0, "q": 0.3, "s": 0.04}
        lens_model = ["NIE_SIMPLE"]
        self.assert_differentials(lens_model, kwargs)

    def test_EPL(self):
        kwargs = {"theta_E": 2.0, "e1": 0.1, "e2": 0.0, "gamma": 2.13}
        lens_model = ["EPL"]
        self.assert_differentials(lens_model, kwargs)

    def test_EPL_Q_PHI(self):
        kwargs = {"theta_E": 2.0, "q": 0.9, "phi": 0.1, "gamma": 2.13}
        lens_model = ["EPL_Q_PHI"]
        self.assert_differentials(lens_model, kwargs)

    def test_EPL_BOXYDISKY(self):
        kwargs = {"theta_E": 2.0, "e1": 0.1, "e2": 0.2, "gamma": 2.13, "a4_a": 0.1}
        lens_model = ["EPL_BOXYDISKY"]
        self.assert_differentials(lens_model, kwargs)

    def test_EPL_numba(self):
        kwargs = {"theta_E": 2.0, "e1": 0.1, "e2": 0.0, "gamma": 2.13}
        lens_model = ["EPL_NUMBA"]
        self.assert_differentials(lens_model, kwargs)

    def test_coreBurk(self):
        kwargs = {"Rs": 2, "alpha_Rs": 1, "r_core": 0.4}
        lens_model = ["coreBURKERT"]
        self.assert_differentials(lens_model, kwargs)
        kwargs = {"Rs": 2, "alpha_Rs": 1, "r_core": 5}
        self.assert_differentials(lens_model, kwargs)

    def test_cnfw(self):
        kwargs = {"Rs": 5.5, "alpha_Rs": 1.0, "r_core": 0.5}
        lens_model = ["CNFW"]
        self.assert_differentials(lens_model, kwargs, potential=True)

    def test_cnfw_ellipse(self):
        kwargs = {"alpha_Rs": 0.1, "Rs": 5.0, "r_core": 0.1, "e1": 0.04, "e2": -0.04}
        lens_model = ["CNFW_ELLIPSE"]
        self.assert_differentials(lens_model, kwargs, potential=True)

    def test_cored_density(self):
        kwargs = {"sigma0": 0.1, "r_core": 8}
        lens_model = ["CORED_DENSITY"]
        self.assert_differentials(lens_model, kwargs)

    def test_cored_density_2(self):
        kwargs = {"sigma0": 0.1, "r_core": 8}
        lens_model = ["CORED_DENSITY_2"]
        self.assert_differentials(lens_model, kwargs)

    def test_cored_density_exp(self):
        kwargs = {"kappa_0": 0.1, "theta_c": 8}
        lens_model = ["CORED_DENSITY_EXP"]
        self.assert_differentials(lens_model, kwargs)

    def test_cored_density_mst(self):
        kwargs = {"lambda_approx": 1.1, "r_core": 8}
        lens_model = ["CORED_DENSITY_MST"]
        self.assert_differentials(lens_model, kwargs)

    def test_cored_density_2_mst(self):
        kwargs = {"lambda_approx": 1.1, "r_core": 8}
        lens_model = ["CORED_DENSITY_2_MST"]
        self.assert_differentials(lens_model, kwargs)

    def test_cored_density_exp_mst(self):
        kwargs = {"lambda_approx": 1.1, "r_core": 8}
        lens_model = ["CORED_DENSITY_EXP_MST"]
        self.assert_differentials(lens_model, kwargs)

    def test_cored_density_uldm_mst(self):
        kwargs = {"lambda_approx": 0.9, "r_core": 8}
        lens_model = ["CORED_DENSITY_ULDM_MST"]
        self.assert_differentials(lens_model, kwargs)

    def test_uldm(self):
        kwargs = {"kappa_0": 0.1, "theta_c": 5, "slope": 7.5}
        lens_model = ["ULDM"]
        self.assert_differentials(lens_model, kwargs)

    def test_const_mag_positive(self):
        kwargs = {"mu_r": 1, "mu_t": 10, "parity": 1, "phi_G": 0.1}
        lens_model = ["CONST_MAG"]
        self.assert_differentials(lens_model, kwargs)

    def test_const_mag_negative(self):
        kwargs = {"mu_r": 1, "mu_t": 10, "parity": -1, "phi_G": 0.1}
        lens_model = ["CONST_MAG"]
        self.assert_differentials(lens_model, kwargs)

    def test_nie_potential(self):
        kwargs = {"theta_E": 2.0, "theta_c": 1.0, "e1": 0.1, "e2": 0.1}
        lens_model = ["NIE_POTENTIAL"]
        self.assert_differentials(lens_model, kwargs)

    def test_multipole(self):
        kwargs = {"m": 4, "a_m": 0.05, "phi_m": 0.1, "center_x": 0.0, "center_y": 0.0}
        lens_model = ["MULTIPOLE"]
        self.assert_differentials(lens_model, kwargs, potential=True)
        kwargs = {"m": 2, "a_m": 0.1, "phi_m": 0.05, "center_x": 0.0, "center_y": 0.0}
        lens_model = ["MULTIPOLE"]
        self.assert_differentials(lens_model, kwargs, potential=True)
        kwargs = {
            "m": 3,
            "a_m": 0.07,
            "phi_m": 0.0,
            "center_x": -0.01,
            "center_y": -0.5,
        }
        lens_model = ["MULTIPOLE"]
        self.assert_differentials(lens_model, kwargs, potential=True)

    def test_elli_slice(self):
        kwargs = {
            "a": 2.0,
            "b": 1.0,
            "psi": 90 * np.pi / 180.0,
            "sigma_0": 5.0,
            "center_x": 1.0,
            "center_y": 0.0,
        }
        lens_model = ["ElliSLICE"]
        self.assert_differentials(lens_model, kwargs, potential=True)

        kwargs = {
            "a": 2.0,
            "b": 1.0,
            "psi": 89 * np.pi / 180.0,
            "sigma_0": 5.0,
            "center_x": 1.0,
            "center_y": 0.01,
        }
        lens_model = ["ElliSLICE"]
        self.assert_differentials(lens_model, kwargs, potential=True)

    def test_curved_arc_const(self):
        kwargs = {
            "tangential_stretch": 4.0,
            "direction": 0.0,
            "curvature": 0.5,
            "center_x": 0,
            "center_y": 0,
        }
        lens_model = ["CURVED_ARC_CONST"]
        self.assert_differentials(lens_model, kwargs, potential=False)
        kwargs = {
            "tangential_stretch": 4.0,
            "direction": 0.5,
            "curvature": 0.5,
            "center_x": 0,
            "center_y": 0,
        }
        self.assert_differentials(lens_model, kwargs, potential=False)

    def test_curved_arc_const_MST(self):
        kwargs = {
            "tangential_stretch": 4.0,
            "radial_stretch": 0.9,
            "direction": 0.0,
            "curvature": 0.5,
            "center_x": 0,
            "center_y": 0,
        }
        lens_model = ["CURVED_ARC_CONST_MST"]
        self.assert_differentials(lens_model, kwargs, potential=False)
        kwargs = {
            "tangential_stretch": 4.0,
            "radial_stretch": 1.1,
            "direction": 0.5,
            "curvature": 0.5,
            "center_x": 0,
            "center_y": 0,
        }
        self.assert_differentials(lens_model, kwargs, potential=False)

    def test_curved_arc_spp(self):
        kwargs = {
            "tangential_stretch": 4.0,
            "radial_stretch": 0.9,
            "direction": 0.5,
            "curvature": 0.5,
            "center_x": 0,
            "center_y": 0,
        }
        lens_model = ["CURVED_ARC_SPP"]
        self.assert_differentials(lens_model, kwargs)

    def test_curved_arc_spt(self):
        kwargs = {
            "tangential_stretch": 4.0,
            "radial_stretch": 0.9,
            "direction": 0.5,
            "curvature": 0.5,
            "center_x": 0,
            "center_y": 0,
            "gamma1": 0.1,
            "gamma2": -0.2,
        }
        lens_model = ["CURVED_ARC_SPT"]
        self.assert_differentials(lens_model, kwargs, potential=False)

    def test_curved_arc_sis_mst(self):
        kwargs = {
            "tangential_stretch": 4.0,
            "radial_stretch": 1,
            "direction": 0.5,
            "curvature": 0.2,
            "center_x": 0,
            "center_y": 0,
        }
        lens_model = ["CURVED_ARC_SIS_MST"]
        self.assert_differentials(lens_model, kwargs)

    def test_curved_arc_tan_diff(self):
        kwargs = {
            "tangential_stretch": 4.0,
            "radial_stretch": 1,
            "direction": 0.5,
            "dtan_dtan": 0.1,
            "curvature": 0.2,
            "center_x": 0,
            "center_y": 0,
        }
        lens_model = ["CURVED_ARC_TAN_DIFF"]
        self.assert_differentials(lens_model, kwargs)

    def test_splcore(self):
        kwargs = {
            "sigma0": 1.0,
            "gamma": 3,
            "r_core": 0.1,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        lens_model = ["SPL_CORE"]
        self.assert_differentials(lens_model, kwargs, potential=False)

        kwargs = {
            "sigma0": 1.0,
            "gamma": 2.0,
            "r_core": 0.1,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        lens_model = ["SPL_CORE"]
        self.assert_differentials(lens_model, kwargs, potential=False)

        kwargs = {
            "sigma0": 1.0,
            "gamma": 2.5,
            "r_core": 0.1,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        lens_model = ["SPL_CORE"]
        self.assert_differentials(lens_model, kwargs, potential=False)

    def test_gnfw(self):
        kwargs = {"alpha_Rs": 1.2, "Rs": 0.8, "gamma_inner": 2.3, "gamma_outer": 3.15}
        lens_model = ["GNFW"]
        self.assert_differentials(lens_model, kwargs, potential=False)

        kwargs = {"alpha_Rs": 1.2, "Rs": 0.8, "gamma_inner": 0.3, "gamma_outer": 3.15}
        lens_model = ["GNFW"]
        self.assert_differentials(lens_model, kwargs, potential=False)

    def test_cse(self):
        kwargs = {"a": 2, "s": 1.0, "e1": 0, "e2": 0, "center_x": 0.0, "center_y": 0.0}
        lens_model = ["CSE"]
        self.assert_differentials(lens_model, kwargs, potential=True)

        kwargs = {
            "a": 2,
            "s": 1.0,
            "e1": 0.3,
            "e2": 0,
            "center_x": 1.0,
            "center_y": 2.0,
        }
        lens_model = ["CSE"]
        self.assert_differentials(lens_model, kwargs, potential=True)

        kwargs = {
            "a": 2,
            "s": 1.0,
            "e1": -0.3,
            "e2": 0,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        lens_model = ["CSE"]
        self.assert_differentials(lens_model, kwargs, potential=True)

        kwargs = {
            "a": 2,
            "s": 1.0,
            "e1": 0.0,
            "e2": 0.3,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        lens_model = ["CSE"]
        self.assert_differentials(lens_model, kwargs, potential=True)

    def test_nfw_cse(self):
        kwargs = {
            "alpha_Rs": 1,
            "Rs": 1,
            "e1": 0,
            "e2": 0,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        lens_model = ["NFW_ELLIPSE_CSE"]
        self.assert_differentials(lens_model, kwargs, potential=True)

        kwargs = {
            "alpha_Rs": 1,
            "Rs": 0.5,
            "e1": 0.2,
            "e2": 0,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        lens_model = ["NFW_ELLIPSE_CSE"]
        self.assert_differentials(lens_model, kwargs, potential=True)

        kwargs = {
            "alpha_Rs": 2,
            "Rs": 0.5,
            "e1": 0,
            "e2": 0.4,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        lens_model = ["NFW_ELLIPSE_CSE"]
        self.assert_differentials(lens_model, kwargs, potential=True)

    def test_hernquist_cse(self):
        kwargs = {
            "sigma0": 1,
            "Rs": 1,
            "e1": 0,
            "e2": 0,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        lens_model = ["HERNQUIST_ELLIPSE_CSE"]
        self.assert_differentials(lens_model, kwargs, potential=True)

        kwargs = {
            "sigma0": 1,
            "Rs": 0.5,
            "e1": 0.2,
            "e2": 0,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        lens_model = ["HERNQUIST_ELLIPSE_CSE"]
        self.assert_differentials(lens_model, kwargs, potential=True)

        kwargs = {
            "sigma0": 2,
            "Rs": 0.5,
            "e1": 0,
            "e2": 0.4,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        lens_model = ["HERNQUIST_ELLIPSE_CSE"]
        self.assert_differentials(lens_model, kwargs, potential=True)

    def test_TNFWC(self):
        kwargs = {"alpha_Rs": 4.0, "Rs": 2.0, "r_core": 0.1, "r_trunc": 200.0}
        lens_model = ["TNFWC"]
        self.assert_differentials(lens_model, kwargs, potential=False)

        kwargs = {"alpha_Rs": 4.0, "Rs": 2.0, "r_core": 0.1, "r_trunc": 5.0}
        lens_model = ["TNFWC"]
        self.assert_differentials(lens_model, kwargs, potential=False)

        kwargs = {"alpha_Rs": 4.0, "Rs": 2.0, "r_core": 0.9, "r_trunc": 5.0}
        lens_model = ["TNFWC"]
        self.assert_differentials(lens_model, kwargs, potential=False)

        kwargs = {"alpha_Rs": 4.0, "Rs": 2.0, "r_core": 2.1, "r_trunc": 5.0}
        lens_model = ["TNFWC"]
        self.assert_differentials(lens_model, kwargs, potential=False)

        kwargs = {"alpha_Rs": 4.0, "Rs": 2.0, "r_core": 12.1, "r_trunc": 5.0}
        lens_model = ["TNFWC"]
        self.assert_differentials(lens_model, kwargs, potential=False)


if __name__ == "__main__":
    pytest.main("-k TestLensModel")



"""Provisional JWST instrument and observational settings.
ZP can be found here : https://jwst-docs.stsci.edu/files/182256933/182256934/1/1669487685625/NRC_ZPs_0995pmap.txt

Sky Brightness needs to be derived from the ETC
"""
import lenstronomy.Util.util as util

__all__ = ["JWST"]


NIRCAM_F200W_band_obs = {
    "exposure_time": 3600.0,
    "sky_brightness": 29.52,  # this is derived using the ETC
    "magnitude_zero_point": 28.00,
    #'detector': 'NRCA1',
    "num_exposures": 1,
    "seeing": None,
    "psf_type": "PIXEL",
}


NIRCAM_F356W_band_obs = {
    "exposure_time": 3600.0,
    "sky_brightness": 28.39,  # this is derived using the ETC
    "magnitude_zero_point": 26.47,
    #'detector': 'NRCALONG',
    "num_exposures": 1,
    "seeing": None,
    "psf_type": "PIXEL",  # note kernel_point_source (the PSF map) must be provided separately
}

# - keyword exposure_time: exposure time per image (in seconds)
# - keyword sky_brightness: sky brightness (in magnitude per square arcseconds in units of electrons)
# - keyword magnitude_zero_point: magnitude in which 1 count (e-) per second per arcsecond square is registered
# - keyword num_exposures: number of exposures that are combined (depends on coadd_years)
# - keyword seeing: Full-Width-at-Half-Maximum (FWHM) of PSF
# - keyword psf_type: string, type of PSF ('GAUSSIAN' and 'PIXEL' supported)


class JWST(object):
    """Class contains JWST instrument and observation configurations."""

    def __init__(self, band="F200W", psf_type="PIXEL", coadd_years=None):
        """

        :param band: string, 'F200W' or 'F356W' supported. Determines obs dictionary.
        :param psf_type: string, type of PSF ('GAUSSIAN', 'PIXEL' supported).
        :param coadd_years: int, number of years corresponding to num_exposures in obs dict. Currently supported: None.
        """

        if band == "F200W":
            self.obs = NIRCAM_F200W_band_obs
            self.arm = "short"
        elif band == "F356W":
            self.obs = NIRCAM_F356W_band_obs
            self.arm = "long"

        else:
            raise ValueError("band %s not supported!" % band)

        if psf_type == "GAUSSIAN":
            self.obs["psf_type"] = "GAUSSIAN"
        elif psf_type != "PIXEL":
            raise ValueError("psf_type %s not supported!" % psf_type)

        if coadd_years is not None:
            raise ValueError(
                " %s coadd_years not supported! "
                "You may manually adjust num_exposures in obs dict if required."
                % coadd_years
            )

        # NIRCAM camera settings
        if self.arm == "short":
            self.camera = {
                "read_noise": 15.77,
                "pixel_scale": 0.031,
                "ccd_gain": 2.05,
            }
        elif self.arm == "long":
            self.camera = {
                "read_noise": 13.25,
                "pixel_scale": 0.063,
                "ccd_gain": 1.82,
            }

        # - keyword read_noise: std of noise generated by read-out (in units of electrons)
        # - keyword pixel_scale: scale (in arcseconds) of pixels
        # - keyword ccd_gain: electrons/ADU (analog-to-digital unit).

    def kwargs_single_band(self):
        """

        :return: merged kwargs from camera and obs dicts
        """
        kwargs = util.merge_dicts(self.camera, self.obs)
        return kwargs



__author__ = "dgilman"


from lenstronomy.LensModel.Profiles.numerical_deflections import TabulatedDeflections
from lenstronomy.LensModel.Profiles.nfw import NFW
from lenstronomy.LensModel.lens_model import LensModel

import numpy as np
import numpy.testing as npt
import pytest


class TestClass:
    def __call__(self, x, y, Rs, norm, center_x=0, center_y=0):
        """NFW profile :param x:

        :param y:
        :param kwargs:
        :return:
        """

        rho0 = self._alpha2rho0(norm, Rs)

        _x = x - center_x
        _y = y - center_y

        R = np.sqrt(_x**2 + _y**2)
        x = R * Rs**-1

        a = 4 * rho0 * Rs * R * self._g(x) / x**2

        cos = _x / R
        sin = _y / R
        return a * cos, a * sin

    def _g(self, X):
        """Analytic solution of integral for NFW profile to compute deflection angel and
        gamma.

        :param x: R/Rs
        :type x: float >0
        """
        c = 0.000001
        if isinstance(X, int) or isinstance(X, float):
            if X < 1:
                x = max(c, X)
                a = np.log(x / 2.0) + 1 / np.sqrt(1 - x**2) * np.arccosh(1.0 / x)
            elif X == 1:
                a = 1 + np.log(1.0 / 2.0)
            else:  # X > 1:
                a = np.log(X / 2) + 1 / np.sqrt(X**2 - 1) * np.arccos(1.0 / X)

        else:
            a = np.empty_like(X)
            X[X <= c] = c
            x = X[X < 1]
            a[X < 1] = np.log(x / 2.0) + 1 / np.sqrt(1 - x**2) * np.arccosh(1.0 / x)
            a[X == 1] = 1 + np.log(1.0 / 2.0)
            x = X[X > 1]
            a[X > 1] = np.log(x / 2) + 1 / np.sqrt(x**2 - 1) * np.arccos(1.0 / x)

        return a

    def _alpha2rho0(self, alpha_Rs, Rs):
        """Convert angle at Rs into rho0."""

        rho0 = alpha_Rs / (4.0 * Rs**2 * (1.0 + np.log(1.0 / 2.0)))
        return rho0


class TestNumericalAlpha(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        self.numerical_alpha = TabulatedDeflections(custom_class=TestClass())
        self.nfw = NFW()

    def test_no_potential(self):
        npt.assert_raises(Exception, self.numerical_alpha.function, 0.0, 0.0, 0.0, 0.0)

    def test_derivatives(self):
        Rs = 10.0
        alpha_Rs = 10
        x = np.linspace(Rs, Rs, 1000)
        y = np.linspace(0.2 * Rs, 2 * Rs, 1000)
        center_x, center_y = -1.2, 0.46

        zlist_single = [0.5, 0.5]
        zlist_multi = [0.5, 0.8]
        zlist = [zlist_single, zlist_multi]
        numerical_alpha_class = TestClass()

        for i, flag in enumerate([False, True]):
            lensmodel = LensModel(
                lens_model_list=["TABULATED_DEFLECTIONS", "NFW"],
                z_source=1.5,
                z_lens=0.5,
                lens_redshift_list=zlist[i],
                multi_plane=flag,
                numerical_alpha_class=numerical_alpha_class,
            )

            lensmodel_nfw = LensModel(
                lens_model_list=["NFW", "NFW"],
                z_source=1.5,
                z_lens=0.5,
                lens_redshift_list=zlist[i],
                multi_plane=flag,
                numerical_alpha_class=numerical_alpha_class,
            )

            keywords_num = [
                {
                    "norm": alpha_Rs,
                    "Rs": Rs,
                    "center_x": center_x,
                    "center_y": center_y,
                },
                {
                    "alpha_Rs": 0.7 * alpha_Rs,
                    "Rs": 2 * Rs,
                    "center_x": center_x,
                    "center_y": center_y,
                },
            ]
            keywords_nfw = [
                {
                    "alpha_Rs": alpha_Rs,
                    "Rs": Rs,
                    "center_x": center_x,
                    "center_y": center_y,
                },
                {
                    "alpha_Rs": 0.7 * alpha_Rs,
                    "Rs": 2 * Rs,
                    "center_x": center_x,
                    "center_y": center_y,
                },
            ]

            dx, dy = lensmodel.alpha(x, y, keywords_num)
            dxnfw, dynfw = lensmodel_nfw.alpha(x, y, keywords_nfw)
            npt.assert_almost_equal(dx, dxnfw)
            npt.assert_almost_equal(dy, dynfw)

    def test_hessian(self):
        Rs = 10.0
        alpha_Rs = 2
        x = np.linspace(Rs, Rs, 1000)
        y = np.linspace(0.2 * Rs, 2 * Rs, 1000)
        center_x, center_y = -1.2, 0.46

        zlist_single = [0.5, 0.5]
        zlist_multi = [0.5, 0.8]
        zlist = [zlist_single, zlist_multi]
        numerical_alpha_class = TestClass()

        for i, flag in enumerate([False, True]):
            lensmodel = LensModel(
                lens_model_list=["TABULATED_DEFLECTIONS", "NFW"],
                z_source=1.5,
                z_lens=0.5,
                lens_redshift_list=zlist[i],
                multi_plane=flag,
                numerical_alpha_class=numerical_alpha_class,
            )

            lensmodel_nfw = LensModel(
                lens_model_list=["NFW", "NFW"],
                z_source=1.5,
                z_lens=0.5,
                lens_redshift_list=zlist[i],
                multi_plane=flag,
                numerical_alpha_class=numerical_alpha_class,
            )

            keywords_num = [
                {
                    "norm": alpha_Rs,
                    "Rs": Rs,
                    "center_x": center_x,
                    "center_y": center_y,
                },
                {
                    "alpha_Rs": 0.7 * alpha_Rs,
                    "Rs": 2 * Rs,
                    "center_x": center_x,
                    "center_y": center_y,
                },
            ]
            keywords_nfw = [
                {
                    "alpha_Rs": alpha_Rs,
                    "Rs": Rs,
                    "center_x": center_x,
                    "center_y": center_y,
                },
                {
                    "alpha_Rs": 0.7 * alpha_Rs,
                    "Rs": 2 * Rs,
                    "center_x": center_x,
                    "center_y": center_y,
                },
            ]

            hess_num = lensmodel.hessian(x, y, keywords_num)
            hess_nfw = lensmodel_nfw.hessian(x, y, keywords_nfw)
            for hn, hnfw in zip(hess_num, hess_nfw):
                diff = hn * hnfw**-1
                L = len(diff)
                npt.assert_almost_equal(np.sum(diff) * L**-1, 1, 6)


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

# description of the polar shapelets in potential space

import numpy as np
import math
import numpy.polynomial.hermite as hermite
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["CartShapelets"]


class CartShapelets(LensProfileBase):
    """This class contains the function and the derivatives of the cartesian
    shapelets."""

    param_names = ["coeffs", "beta", "center_x", "center_y"]
    lower_limit_default = {"coeffs": [0], "beta": 0, "center_x": -100, "center_y": -100}
    upper_limit_default = {
        "coeffs": [100],
        "beta": 100,
        "center_x": 100,
        "center_y": 100,
    }

    def function(self, x, y, coeffs, beta, center_x=0, center_y=0):
        shapelets = self._createShapelet(coeffs)
        n_order = self._get_num_n(len(coeffs))
        n = len(np.atleast_1d(x))
        if n <= 1:
            f_ = self._shapeletOutput(x, y, beta, shapelets, precalc=False)
        else:
            H_x, H_y = self.pre_calc(x, y, beta, n_order, center_x, center_y)
            f_ = self._shapeletOutput(H_x, H_y, beta, shapelets)
        return f_

    def derivatives(self, x, y, coeffs, beta, center_x=0, center_y=0):
        """Returns df/dx and df/dy of the function."""
        shapelets = self._createShapelet(coeffs)
        n_order = self._get_num_n(len(coeffs))
        dx_shapelets = self._dx_shapelets(shapelets, beta)
        dy_shapelets = self._dy_shapelets(shapelets, beta)
        n = len(np.atleast_1d(x))
        if n <= 1:
            f_x = self._shapeletOutput(x, y, beta, dx_shapelets, precalc=False)
            f_y = self._shapeletOutput(x, y, beta, dy_shapelets, precalc=False)
        else:
            H_x, H_y = self.pre_calc(x, y, beta, n_order + 1, center_x, center_y)
            f_x = self._shapeletOutput(H_x, H_y, beta, dx_shapelets)
            f_y = self._shapeletOutput(H_x, H_y, beta, dy_shapelets)
        return f_x, f_y

    def hessian(self, x, y, coeffs, beta, center_x=0, center_y=0):
        """Returns Hessian matrix of function d^2f/dx^2, d^2/dxdy, d^2/dydx,
        d^f/dy^2."""
        shapelets = self._createShapelet(coeffs)
        n_order = self._get_num_n(len(coeffs))
        dxx_shapelets = self._dxx_shapelets(shapelets, beta)
        dyy_shapelets = self._dyy_shapelets(shapelets, beta)
        dxy_shapelets = self._dxy_shapelets(shapelets, beta)
        n = len(np.atleast_1d(x))
        if n <= 1:
            f_xx = self._shapeletOutput(x, y, beta, dxx_shapelets, precalc=False)
            f_yy = self._shapeletOutput(x, y, beta, dyy_shapelets, precalc=False)
            f_xy = self._shapeletOutput(x, y, beta, dxy_shapelets, precalc=False)
        else:
            H_x, H_y = self.pre_calc(x, y, beta, n_order + 2, center_x, center_y)
            f_xx = self._shapeletOutput(H_x, H_y, beta, dxx_shapelets)
            f_yy = self._shapeletOutput(H_x, H_y, beta, dyy_shapelets)
            f_xy = self._shapeletOutput(H_x, H_y, beta, dxy_shapelets)
        return f_xx, f_xy, f_xy, f_yy

    def _createShapelet(self, coeffs):
        """Returns a shapelet array out of the coefficients *a, up to order l.

        :param num_l: order of shapelets
        :type num_l: int.
        :param coeff: shapelet coefficients
        :type coeff: floats
        :returns: complex array
        :raises: AttributeError, KeyError
        """
        n_coeffs = len(coeffs)
        num_n = self._get_num_n(n_coeffs)
        shapelets = np.zeros((num_n + 1, num_n + 1))
        n = 0
        k = 0
        for coeff in coeffs:
            shapelets[n - k][k] = coeff
            k += 1
            if k == n + 1:
                n += 1
                k = 0
        return shapelets

    def _shapeletOutput(self, x, y, beta, shapelets, precalc=True):
        """Returns the numerical values of a set of shapelets at polar coordinates.

        :param shapelets: set of shapelets [l=,r=,a_lr=]
        :type shapelets: array of size (n,3)
        :param coordPolar: set of coordinates in polar units
        :type coordPolar: array of size (n,2)
        :returns: array of same size with coords [r,phi]
        :raises: AttributeError, KeyError
        """
        n = len(np.atleast_1d(x))
        if n <= 1:
            values = 0.0
        else:
            values = np.zeros(len(x[0]))
        n = 0
        k = 0
        i = 0
        num_n = len(shapelets)
        while i < num_n * (num_n + 1) / 2:
            values += self._function(
                x, y, shapelets[n - k][k], beta, n - k, k, precalc=precalc
            )
            k += 1
            if k == n + 1:
                n += 1
                k = 0
            i += 1
        return values

    def _function(self, x, y, amp, beta, n1, n2, center_x=0, center_y=0, precalc=False):
        """

        :param amp: amplitude of shapelet
        :param beta: scale factor of shapelet
        :param n1: x-order
        :param n2: y-order
        :param center_x: center in x
        :param center_y: center in y
        :return:
        """
        if precalc:
            return amp * x[n1] * y[n2] / beta
        x_ = x - center_x
        y_ = y - center_y
        return amp * self.phi_n(n1, x_ / beta) * self.phi_n(n2, y_ / beta) / beta

    def _dx_shapelets(self, shapelets, beta):
        """Computes the derivative d/dx of the shapelet coeffs :param shapelets:

        :param beta:
        :return:
        """
        num_n = len(shapelets)
        dx = np.zeros((num_n + 1, num_n + 1))
        for n1 in range(num_n):
            for n2 in range(num_n):
                amp = shapelets[n1][n2]
                dx[n1 + 1][n2] -= np.sqrt((n1 + 1) / 2.0) * amp
                if n1 > 0:
                    dx[n1 - 1][n2] += np.sqrt(n1 / 2.0) * amp
        return dx / beta

    def _dy_shapelets(self, shapelets, beta):
        """Computes the derivative d/dx of the shapelet coeffs :param shapelets:

        :param beta:
        :return:
        """
        num_n = len(shapelets)
        dy = np.zeros((num_n + 1, num_n + 1))
        for n1 in range(num_n):
            for n2 in range(num_n):
                amp = shapelets[n1][n2]
                dy[n1][n2 + 1] -= np.sqrt((n2 + 1) / 2.0) * amp
                if n2 > 0:
                    dy[n1][n2 - 1] += np.sqrt(n2 / 2.0) * amp
        return dy / beta

    def _dxx_shapelets(self, shapelets, beta):
        dx_shapelets = self._dx_shapelets(shapelets, beta)
        return self._dx_shapelets(dx_shapelets, beta)

    def _dyy_shapelets(self, shapelets, beta):
        dy_shapelets = self._dy_shapelets(shapelets, beta)
        return self._dy_shapelets(dy_shapelets, beta)

    def _dxy_shapelets(self, shapelets, beta):
        dy_shapelets = self._dy_shapelets(shapelets, beta)
        return self._dx_shapelets(dy_shapelets, beta)

    def H_n(self, n, x):
        """Constructs the Hermite polynomial of order n at position x (dimensionless)

        :param n: The n'the basis function.
        :type name: int.
        :param x: 1-dim position (dimensionless)
        :type state: float or numpy array.
        :returns: array-- H_n(x).
        :raises: AttributeError, KeyError
        """
        n_array = np.zeros(n + 1)
        n_array[n] = 1
        return hermite.hermval(
            x, n_array, tensor=False
        )  # attention, this routine calculates every single hermite polynomial and multiplies it with zero (exept the right one)

    def phi_n(self, n, x):
        """Constructs the 1-dim basis function (formula (1) in Refregier et al. 2001)

        :param n: The n'the basis function.
        :type name: int.
        :param x: 1-dim position (dimensionless)
        :type state: float or numpy array.
        :returns: array-- phi_n(x).
        :raises: AttributeError, KeyError
        """
        prefactor = 1.0 / np.sqrt(2**n * np.sqrt(np.pi) * math.factorial(n))
        return prefactor * self.H_n(n, x) * np.exp(-(x**2) / 2.0)

    def pre_calc(self, x, y, beta, n_order, center_x, center_y):
        """Calculates the H_n(x) and H_n(y) for a given x-array and y-array :param x:

        :param y:
        :param amp:
        :param beta:
        :param n_order:
        :param center_x:
        :param center_y:
        :return: list of H_n(x) and H_n(y)
        """

        n = len(np.atleast_1d(x))
        x_ = x - center_x
        y_ = y - center_y
        H_x = np.empty((n_order + 1, n))
        H_y = np.empty((n_order + 1, n))
        for n in range(0, n_order + 1):
            prefactor = 1.0 / np.sqrt(2**n * np.sqrt(np.pi) * math.factorial(n))
            n_array = np.zeros(n + 1)
            n_array[n] = 1
            H_x[n] = (
                hermite.hermval(x_ / beta, n_array)
                * prefactor
                * np.exp(-((x_ / beta) ** 2) / 2.0)
            )
            H_y[n] = (
                hermite.hermval(y_ / beta, n_array)
                * prefactor
                * np.exp(-((y_ / beta) ** 2) / 2.0)
            )
        return H_x, H_y

    def _get_num_n(self, n_coeffs):
        """

        :param n_coeffs: number of coeffs
        :return: number of n_l of order of the shapelets
        """
        num_n = round((math.sqrt(8 * n_coeffs + 1) - 1) / 2.0 + 0.499)
        return int(num_n)



import pytest
from lenstronomy.LightModel.Profiles.moffat import Moffat


class TestMoffat(object):
    """Class to test the Moffat profile."""

    def setup_method(self):
        pass

    def test_function(self):
        """

        :return:
        """
        profile = Moffat()
        output = profile.function(
            x=1.0, y=1.0, amp=1.0, alpha=2.0, beta=1.0, center_x=0, center_y=0
        )
        assert output == 0.6666666666666666


if __name__ == "__main__":
    pytest.main()



from lenstronomy.PointSource.Types.source_position import SourcePositions
from lenstronomy.LensModel.lens_model import LensModel
import pytest
import numpy.testing as npt


class TestLensedPosition(object):
    def setup_method(self):
        lens_model = LensModel(lens_model_list=["SIS"])
        self.kwargs_lens = [{"theta_E": 1, "center_x": 0, "center_y": 0}]
        self.ps_mag = SourcePositions(lens_model=lens_model, fixed_magnification=True)
        self.ps = SourcePositions(lens_model=lens_model, fixed_magnification=False)
        self.kwargs = {"point_amp": [2, 1], "ra_source": 0.1, "dec_source": 0}
        self.kwargs_mag = {"source_amp": 1, "ra_source": 0.1, "dec_source": 0}

    def test_image_position(self):
        x_img, y_img = self.ps.image_position(self.kwargs, self.kwargs_lens)

        lens_model = LensModel(lens_model_list=["SIS"])
        x_src, y_src = lens_model.ray_shooting(x_img, y_img, kwargs=self.kwargs_lens)
        npt.assert_almost_equal(x_src, self.kwargs["ra_source"])
        npt.assert_almost_equal(y_src, self.kwargs["dec_source"])

    def test_source_position(self):
        x_src, y_src = self.ps.source_position(
            self.kwargs, kwargs_lens=self.kwargs_lens
        )
        npt.assert_almost_equal(x_src, self.kwargs["ra_source"])
        npt.assert_almost_equal(y_src, self.kwargs["dec_source"])

    def test_image_amplitude(self):
        x_img, y_img = self.ps.image_position(self.kwargs, kwargs_lens=self.kwargs_lens)
        amp = self.ps_mag.image_amplitude(
            self.kwargs_mag,
            kwargs_lens=self.kwargs_lens,
            x_pos=None,
            y_pos=None,
            magnification_limit=None,
            kwargs_lens_eqn_solver=None,
        )
        amp_pos = self.ps_mag.image_amplitude(
            self.kwargs_mag,
            kwargs_lens=self.kwargs_lens,
            x_pos=x_img,
            y_pos=y_img,
            magnification_limit=None,
            kwargs_lens_eqn_solver=None,
        )
        npt.assert_almost_equal(amp_pos, amp)

        amp = self.ps.image_amplitude(
            self.kwargs,
            kwargs_lens=self.kwargs_lens,
            x_pos=x_img,
            y_pos=y_img,
            magnification_limit=None,
            kwargs_lens_eqn_solver=None,
        )
        npt.assert_almost_equal(amp, self.kwargs["point_amp"])

        # see if works with mag_pert defined
        self.kwargs["mag_pert"] = [0.1, 0.1]
        amp_pert = self.ps.image_amplitude(
            self.kwargs,
            kwargs_lens=self.kwargs_lens,
            x_pos=x_img,
            y_pos=y_img,
            magnification_limit=None,
            kwargs_lens_eqn_solver=None,
        )
        npt.assert_almost_equal(amp_pert, 0.1 * amp)

    def test_source_amplitude(self):
        amp = self.ps.source_amplitude(self.kwargs, kwargs_lens=self.kwargs_lens)
        amp_mag = self.ps_mag.source_amplitude(
            self.kwargs_mag, kwargs_lens=self.kwargs_lens
        )
        npt.assert_almost_equal(amp_mag, self.kwargs_mag["source_amp"])
        assert amp != amp_mag


if __name__ == "__main__":
    pytest.main()



from scipy import fftpack, ndimage, signal
import numpy as np
import threading
import lenstronomy.Util.kernel_util as kernel_util
import lenstronomy.Util.util as util
import lenstronomy.Util.image_util as image_util

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()

_rfft_mt_safe = True  # (NumpyVersion(np.__version__) >= '1.9.0.dev-e24486e')
_rfft_lock = threading.Lock()


def _centered(arr, newshape):
    # Return the center newshape portion of the array.
    newshape = np.asarray(newshape)
    currshape = np.array(arr.shape)
    startind = (currshape - newshape) // 2
    endind = startind + newshape
    myslice = [slice(startind[k], endind[k]) for k in range(len(endind))]
    return arr[tuple(myslice)]


@export
class PixelKernelConvolution(object):
    """Class to compute convolutions for a given pixelized kernel (fft, grid)"""

    def __init__(self, kernel, convolution_type="fft_static"):
        """

        :param kernel: 2d array, convolution kernel
        :param convolution_type: string, 'fft', 'grid', 'fft_static' mode of 2d convolution
        """
        self._kernel = kernel
        if convolution_type not in ["fft", "grid", "fft_static"]:
            raise ValueError("convolution_type %s not supported!" % convolution_type)
        self._type = convolution_type
        self._pre_computed = False

    def pixel_kernel(self, num_pix=None):
        """Access pixelated kernel.

        :param num_pix: size of returned kernel (odd number per axis). If None, return
            the original kernel.
        :return: pixel kernel centered
        """
        if num_pix is not None:
            return kernel_util.cut_psf(self._kernel, num_pix)
        return self._kernel

    def copy_transpose(self):
        """

        :return: copy of the class with kernel set to the transpose of original one
        """
        return PixelKernelConvolution(self._kernel.T, convolution_type=self._type)

    def convolution2d(self, image):
        """

        :param image: 2d array (image) to be convolved
        :return: fft convolution
        """
        if self._type == "fft":
            image_conv = signal.fftconvolve(image, self._kernel, mode="same")
        elif self._type == "fft_static":
            image_conv = self._static_fft(image, mode="same")
        elif self._type == "grid":
            image_conv = signal.convolve2d(image, self._kernel, mode="same")
        else:
            raise ValueError("convolution_type %s not supported!" % self._type)
        return image_conv

    def _static_fft(self, image, mode="same"):
        """Scipy fft convolution with saved static fft kernel.

        :param image: 2d numpy array to be convolved
        :return:
        """
        in1 = image
        in1 = np.asarray(in1)
        if self._pre_computed is False:
            (
                self._s1,
                self._s2,
                self._complex_result,
                self._shape,
                self._fshape,
                self._fslice,
                self._sp2,
            ) = self._static_pre_compute(image)
            self._pre_computed = True
        s1, s2, complex_result, shape, fshape, fslice, sp2 = (
            self._s1,
            self._s2,
            self._complex_result,
            self._shape,
            self._fshape,
            self._fslice,
            self._sp2,
        )
        # if in1.ndim == in2.ndim == 0:  # scalar inputs
        #    return in1 * in2
        # elif not in1.ndim == in2.ndim:
        #    raise ValueError("in1 and in2 should have the same dimensionality")
        # elif in1.size == 0 or in2.size == 0:  # empty arrays
        #    return np.array([])

        # Check that input sizes are compatible with 'valid' mode
        # if _inputs_swap_needed(mode, s1, s2):
        # Convolution is commutative; order doesn't have any effect on output
        # only applicable for 'valid' mode
        #    in1, s1, in2, s2 = in2, s2, in1, s1

        # Pre-1.9 NumPy FFT routines are not threadsafe.  For older NumPys, make
        # sure we only call rfftn/irfftn from one thread at a time.
        if not complex_result and (_rfft_mt_safe or _rfft_lock.acquire(False)):
            try:
                sp1 = np.fft.rfftn(in1, fshape)
                ret = np.fft.irfftn(sp1 * sp2, fshape)[fslice].copy()
            finally:
                if not _rfft_mt_safe:
                    _rfft_lock.release()
        else:
            # If we're here, it's either because we need a complex result, or we
            # failed to acquire _rfft_lock (meaning rfftn isn't threadsafe and
            # is already in use by another thread).  In either case, use the
            # (threadsafe but slower) SciPy complex-FFT routines instead.
            sp1 = fftpack.fftn(in1, fshape)
            ret = fftpack.ifftn(sp1 * sp2)[fslice].copy()
            if not complex_result:
                ret = ret.real

        if mode == "full":
            return ret
        elif mode == "same":
            return _centered(ret, s1)
        elif mode == "valid":
            return _centered(ret, s1 - s2 + 1)
        else:
            raise ValueError("Acceptable mode flags are 'valid'," " 'same', or 'full'.")

    def _static_pre_compute(self, image):
        """Pre-compute Fourier transformed kernel and shape quantities to speed up
        convolution.

        :param image: 2d numpy array
        :return:
        """
        in1 = image
        in2 = self._kernel
        s1 = np.array(in1.shape)
        s2 = np.array(in2.shape)
        complex_result = np.issubdtype(in1.dtype, np.complexfloating) or np.issubdtype(
            in2.dtype, np.complexfloating
        )
        shape = s1 + s2 - 1

        # Check that input sizes are compatible with 'valid' mode
        # if _inputs_swap_needed(mode, s1, s2):
        # Convolution is commutative; order doesn't have any effect on output
        # only applicable for 'valid' mode
        #    in1, s1, in2, s2 = in2, s2, in1, s1

        # Speed up FFT by padding to optimal size for FFTPACK
        fshape = [fftpack.next_fast_len(int(d)) for d in shape]
        fslice = tuple([slice(0, int(sz)) for sz in shape])
        # Pre-1.9 NumPy FFT routines are not threadsafe.  For older NumPys, make
        # sure we only call rfftn/irfftn from one thread at a time.
        if not complex_result and (_rfft_mt_safe or _rfft_lock.acquire(False)):
            try:
                sp2 = np.fft.rfftn(in2, fshape)
            finally:
                if not _rfft_mt_safe:
                    _rfft_lock.release()
        else:
            # If we're here, it's either because we need a complex result, or we
            # failed to acquire _rfft_lock (meaning rfftn isn't threadsafe and
            # is already in use by another thread).  In either case, use the
            # (threadsafe but slower) SciPy complex-FFT routines instead.
            sp2 = fftpack.fftn(in2, fshape)
        return s1, s2, complex_result, shape, fshape, fslice, sp2

    def re_size_convolve(self, image_low_res, image_high_res=None):
        """

        :param image_low_res: regular sampled image/model
        :param image_high_res: supersampled image/model to be convolved on a regular pixel grid
        :return: convolved and re-sized image
        """
        return self.convolution2d(image_low_res)


@export
class SubgridKernelConvolution(object):
    """Class to compute the convolution on a supersampled grid with partial convolution
    computed on the regular grid."""

    def __init__(
        self,
        kernel_supersampled,
        supersampling_factor,
        supersampling_kernel_size=None,
        convolution_type="fft_static",
    ):
        """

        :param kernel_supersampled: kernel in supersampled pixels
        :param supersampling_factor: supersampling factor relative to the image pixel grid
        :param supersampling_kernel_size: number of pixels (in units of the image pixels) that are convolved with the
         supersampled kernel
        """
        # n_high = len(kernel_supersampled)
        self._supersampling_factor = supersampling_factor
        # numPix = int(n_high / self._supersampling_factor)
        # if self._supersampling_factor % 2 == 0:
        #    self._kernel = kernel_util.averaging_even_kernel(kernel_supersampled, self._supersampling_factor)
        # else:
        #    self._kernel = util.averaging(kernel_supersampled, numGrid=n_high, numPix=numPix)
        if supersampling_kernel_size is None:
            kernel_low_res, kernel_high_res = np.zeros((3, 3)), kernel_supersampled
            self._low_res_convolution = False
        else:
            kernel_low_res, kernel_high_res = kernel_util.split_kernel(
                kernel_supersampled,
                supersampling_kernel_size,
                self._supersampling_factor,
            )
            self._low_res_convolution = True
        self._low_res_conv = PixelKernelConvolution(
            kernel_low_res, convolution_type=convolution_type
        )
        self._high_res_conv = PixelKernelConvolution(
            kernel_high_res, convolution_type=convolution_type
        )

    def convolution2d(self, image):
        """

        :param image: 2d array (high resoluton image) to be convolved and re-sized
        :return: convolved image
        """

        image_high_res_conv = self._high_res_conv.convolution2d(image)
        image_resized_conv = image_util.re_size(
            image_high_res_conv, self._supersampling_factor
        )
        if self._low_res_convolution is True:
            image_resized = image_util.re_size(image, self._supersampling_factor)
            image_resized_conv += self._low_res_conv.convolution2d(image_resized)
        return image_resized_conv

    def re_size_convolve(self, image_low_res, image_high_res):
        """

        :param image_high_res: supersampled image/model to be convolved on a regular pixel grid
        :return: convolved and re-sized image
        """
        image_high_res_conv = self._high_res_conv.convolution2d(image_high_res)
        image_resized_conv = image_util.re_size(
            image_high_res_conv, self._supersampling_factor
        )
        if self._low_res_convolution is True:
            image_resized_conv += self._low_res_conv.convolution2d(image_low_res)
        return image_resized_conv


@export
class MultiGaussianConvolution(object):
    """Class to perform a convolution consisting of multiple 2d Gaussians This is aimed
    to lead to a speed-up without significant loss of accuracy do to the simplified
    convolution kernel relative to a pixelized kernel."""

    def __init__(
        self,
        sigma_list,
        fraction_list,
        pixel_scale,
        supersampling_factor=1,
        supersampling_convolution=False,
        truncation=2,
    ):
        """

        :param sigma_list: list of std value of Gaussian kernel
        :param fraction_list: fraction of flux to be convoled with each Gaussian kernel
        :param pixel_scale: scale of pixel width (to convert sigmas into units of pixels)
        :param truncation: float. Truncate the filter at this many standard deviations.
         Default is 4.0.
        """
        self._num_gaussians = len(sigma_list)
        self._sigmas_scaled = np.array(sigma_list) / pixel_scale
        if supersampling_convolution is True:
            self._sigmas_scaled *= supersampling_factor
        self._fraction_list = fraction_list / np.sum(fraction_list)
        assert len(self._sigmas_scaled) == len(self._fraction_list)
        self._truncation = truncation
        self._pixel_scale = pixel_scale
        self._supersampling_factor = supersampling_factor
        self._supersampling_convolution = supersampling_convolution

    def convolution2d(self, image):
        """2d convolution.

        :param image: 2d numpy array, image to be convolved
        :return: convolved image, 2d numpy array
        """
        image_conv = None
        for i in range(self._num_gaussians):
            if image_conv is None:
                image_conv = (
                    ndimage.gaussian_filter(
                        image,
                        self._sigmas_scaled[i],
                        mode="nearest",
                        truncate=self._truncation,
                    )
                    * self._fraction_list[i]
                )
            else:
                image_conv += (
                    ndimage.gaussian_filter(
                        image,
                        self._sigmas_scaled[i],
                        mode="nearest",
                        truncate=self._truncation,
                    )
                    * self._fraction_list[i]
                )
        return image_conv

    def re_size_convolve(self, image_low_res, image_high_res):
        """

        :param image_high_res: supersampled image/model to be convolved on a regular pixel grid
        :return: convolved and re-sized image
        """
        if self._supersampling_convolution is True:
            image_high_res_conv = self.convolution2d(image_high_res)
            image_resized_conv = image_util.re_size(
                image_high_res_conv, self._supersampling_factor
            )
        else:
            image_resized_conv = self.convolution2d(image_low_res)
        return image_resized_conv

    def pixel_kernel(self, num_pix):
        """Computes a pixelized kernel from the MGE parameters.

        :param num_pix: int, size of kernel (odd number per axis)
        :return: pixel kernel centered
        """
        from lenstronomy.LightModel.Profiles.gaussian import MultiGaussian

        mg = MultiGaussian()
        x, y = util.make_grid(numPix=num_pix, deltapix=self._pixel_scale)
        kernel = mg.function(x, y, amp=self._fraction_list, sigma=self._sigmas_scaled)
        kernel = util.array2image(kernel)
        return kernel / np.sum(kernel)


@export
class FWHMGaussianConvolution(object):
    """Uses a two-dimensional Gaussian function with same FWHM of given kernel as
    approximation."""

    def __init__(self, kernel, truncation=4):
        """

        :param kernel: 2d kernel
        :param truncation: sigma scaling of kernel truncation
        """
        fwhm = kernel_util.fwhm_kernel(kernel)
        self._sigma = util.fwhm2sigma(fwhm)
        self._truncation = truncation

    def convolution2d(self, image):
        """2d convolution.

        :param image: 2d numpy array, image to be convolved
        :return: convolved image, 2d numpy array
        """

        image_conv = ndimage.filters.gaussian_filter(
            image, self._sigma, mode="nearest", truncate=self._truncation
        )
        return image_conv


@export
class MGEConvolution(object):
    """Approximates a 2d kernel with an azimuthal Multi-Gaussian expansion."""

    def __init__(self, kernel, pixel_scale, order=1):
        """

        :param kernel: 2d convolution kernel (centered, odd axis number)
        :param order: order of Multi-Gaussian Expansion
        """
        amps, sigmas, norm = kernel_util.mge_kernel(kernel, order=order)
        # make instance o MultiGaussian convolution kernel
        self._mge_conv = MultiGaussianConvolution(
            sigma_list=sigmas * pixel_scale,
            fraction_list=np.array(amps) / np.sum(amps),
            pixel_scale=pixel_scale,
            truncation=4,
        )
        self._kernel = kernel
        # store difference between MGE approximation and real kernel

    def convolution2d(self, image):
        """

        :param image:
        :return:
        """
        return self._mge_conv.convolution2d(image)

    def kernel_difference(self):
        """

        :return: difference between true kernel and MGE approximation
        """
        kernel_mge = self._mge_conv.pixel_kernel(num_pix=len(self._kernel))
        return self._kernel - kernel_mge



from lenstronomy.LightModel.light_model import LightModel

__all__ = ["LightParam"]


class LightParam(object):
    """Class manages the parameters corresponding to the LightModel() module.

    Also manages linear parameter handling.
    """

    def __init__(
        self,
        light_model_list,
        kwargs_fixed,
        kwargs_lower=None,
        kwargs_upper=None,
        param_type="light",
        linear_solver=True,
    ):
        """

        :param light_model_list: list of light models
        :param kwargs_fixed: list of keyword arguments corresponding to parameters held fixed during sampling
        :param kwargs_lower: list of keyword arguments indicating hard lower limit of the parameter space
        :param kwargs_upper: list of keyword arguments indicating hard upper limit of the parameter space
        :param param_type: string (optional), adding specifications in the output strings (such as lens light or
         source light)
        :param linear_solver: bool, if True fixes the linear amplitude parameters 'amp' (avoid sampling) such that they
         get overwritten by the linear solver solution.
        """
        self._lightModel = LightModel(light_model_list=light_model_list)
        self._param_name_list = self._lightModel.param_name_list
        self._type = param_type
        self.model_list = light_model_list
        self.kwargs_fixed = kwargs_fixed
        if linear_solver:
            self.kwargs_fixed = self._lightModel.add_fixed_linear(self.kwargs_fixed)
        self._linear_solve = linear_solver
        if kwargs_lower is None:
            kwargs_lower = []
            for func in self._lightModel.func_list:
                kwargs_lower.append(func.lower_limit_default)
        if kwargs_upper is None:
            kwargs_upper = []
            for func in self._lightModel.func_list:
                kwargs_upper.append(func.upper_limit_default)
        self.lower_limit = kwargs_lower
        self.upper_limit = kwargs_upper
        # check that n_max is fixed
        for k, model in enumerate(self.model_list):
            if model in ["SHAPELETS", "SHAPELETS_POLAR", "SHAPELETS_POLAR_EXP"]:
                if "n_max" not in self.kwargs_fixed[k]:
                    Warning("n_max needs to be fixed in %s." % model)

    @property
    def param_name_list(self):
        return self._param_name_list

    def get_params(self, args, i):
        """

        :param args: list of floats corresponding ot the arguments being sampled
        :param i: int, index of the first argument that is managed/read-out by this class
        :return: keyword argument list of the light profile, index after reading out the arguments corresponding to
         this class
        """
        kwargs_list = []
        for k, model in enumerate(self.model_list):
            kwargs = {}
            kwargs_fixed = self.kwargs_fixed[k]
            param_names = self._param_name_list[k]
            for name in param_names:
                if name not in kwargs_fixed:
                    if (
                        model in ["SHAPELETS", "SHAPELETS_POLAR", "SHAPELETS_POLAR_EXP"]
                        and name == "amp"
                    ):
                        if "n_max" in kwargs_fixed:
                            n_max = kwargs_fixed["n_max"]
                        else:
                            raise ValueError("n_max needs to be fixed in %s." % model)
                        if model in ["SHAPELETS_POLAR_EXP"]:
                            num_param = int((n_max + 1) ** 2)
                        else:
                            num_param = int((n_max + 1) * (n_max + 2) / 2)
                        kwargs["amp"] = args[i : i + num_param]
                        i += num_param
                    elif (
                        model in ["MULTI_GAUSSIAN", "MULTI_GAUSSIAN_ELLIPSE"]
                        and name == "amp"
                    ):
                        if "sigma" in kwargs_fixed:
                            num_param = len(kwargs_fixed["sigma"])
                        else:
                            raise ValueError("sigma needs to be fixed in %s." % model)
                        kwargs["amp"] = args[i : i + num_param]
                        i += num_param
                    elif (
                        model in ["SLIT_STARLETS", "SLIT_STARLETS_GEN2"]
                        and name == "amp"
                    ):
                        if "n_scales" in kwargs_fixed and "n_pixels" in kwargs_fixed:
                            n_scales = kwargs_fixed["n_scales"]
                            n_pixels = kwargs_fixed["n_pixels"]
                        else:
                            raise ValueError(
                                "'n_scales' and 'n_pixels' both need to be fixed in %s."
                                % model
                            )
                        num_param = n_scales * n_pixels
                        kwargs["amp"] = args[i : i + num_param]
                        i += num_param
                    else:
                        kwargs[name] = args[i]
                        i += 1
                else:
                    kwargs[name] = kwargs_fixed[name]

            kwargs_list.append(kwargs)
        return kwargs_list, i

    def set_params(self, kwargs_list):
        """

        :param kwargs_list: list of keyword arguments of the light profile (free parameter as well as optionally the
         fixed ones)
        :return: list of floats corresponding to the free parameters
        """
        args = []
        for k, model in enumerate(self.model_list):
            kwargs = kwargs_list[k]
            kwargs_fixed = self.kwargs_fixed[k]
            param_names = self._param_name_list[k]
            for name in param_names:
                if name not in kwargs_fixed:
                    if (
                        model in ["SHAPELETS", "SHAPELETS_POLAR", "SHAPELETS_POLAR_EXP"]
                        and name == "amp"
                    ):
                        n_max = kwargs_fixed.get("n_max", kwargs["n_max"])
                        if model in ["SHAPELETS_POLAR_EXP"]:
                            num_param = int((n_max + 1) ** 2)
                        else:
                            num_param = int((n_max + 1) * (n_max + 2) / 2)
                        for i in range(num_param):
                            args.append(kwargs[name][i])
                    elif (
                        model in ["SLIT_STARLETS", "SLIT_STARLETS_GEN2"]
                        and name == "amp"
                    ):
                        if "n_scales" in kwargs_fixed:
                            n_scales = kwargs_fixed["n_scales"]
                        else:
                            raise ValueError(
                                "'n_scales' for SLIT_STARLETS not found in kwargs_fixed"
                            )
                        if "n_pixels" in kwargs_fixed:
                            n_pixels = kwargs_fixed["n_pixels"]
                        else:
                            raise ValueError(
                                "'n_pixels' for SLIT_STARLETS not found in kwargs_fixed"
                            )
                        num_param = n_scales * n_pixels
                        for i in range(num_param):
                            args.append(kwargs[name][i])
                    elif model in ["SLIT_STARLETS", "SLIT_STARLETS_GEN2"] and name in [
                        "n_scales",
                        "n_pixels",
                        "scale",
                        "center_x",
                        "center_y",
                    ]:
                        raise ValueError(
                            "'{}' must be a fixed keyword argument for STARLETS-like models".format(
                                name
                            )
                        )
                    elif (
                        model in ["MULTI_GAUSSIAN", "MULTI_GAUSSIAN_ELLIPSE"]
                        and name == "amp"
                    ):
                        num_param = len(kwargs["sigma"])
                        for i in range(num_param):
                            args.append(kwargs[name][i])
                    elif (
                        model in ["MULTI_GAUSSIAN", "MULTI_GAUSSIAN_ELLIPSE"]
                        and name == "sigma"
                    ):
                        raise ValueError(
                            "'sigma' must be a fixed keyword argument for MULTI_GAUSSIAN"
                        )
                    else:
                        args.append(kwargs[name])
        return args

    def num_param(self, latex_style=False):
        """
        :param latex_style: boolena; if True, returns latex strings for plotting
        :return: int, list of strings with param names
        """
        num = 0
        name_list = []
        for k, model in enumerate(self.model_list):
            kwargs_fixed = self.kwargs_fixed[k]
            param_names = self._param_name_list[k]
            for name in param_names:
                if name not in kwargs_fixed:
                    if (
                        model in ["SHAPELETS", "SHAPELETS_POLAR", "SHAPELETS_POLAR_EXP"]
                        and name == "amp"
                    ):
                        if "n_max" not in kwargs_fixed:
                            raise ValueError(
                                "n_max needs to be fixed in this configuration!"
                            )
                        n_max = kwargs_fixed["n_max"]
                        if model in ["SHAPELETS_POLAR_EXP"]:
                            num_param = int((n_max + 1) ** 2)
                        else:
                            num_param = int((n_max + 1) * (n_max + 2) / 2)
                        num += num_param
                        for i in range(num_param):
                            name_list.append(str(name + "_" + self._type + str(k)))
                    elif (
                        model in ["SLIT_STARLETS", "SLIT_STARLETS_GEN2"]
                        and name == "amp"
                    ):
                        if (
                            "n_scales" not in kwargs_fixed
                            or "n_pixels" not in kwargs_fixed
                        ):
                            raise ValueError(
                                "n_scales and n_pixels need to be fixed when using STARLETS-like models!"
                            )
                        n_scales = kwargs_fixed["n_scales"]
                        n_pixels = kwargs_fixed["n_pixels"]
                        num_param = n_scales * n_pixels
                        num += num_param
                        for i in range(num_param):
                            name_list.append(str(name + "_" + self._type + str(k)))
                    elif (
                        model in ["MULTI_GAUSSIAN", "MULTI_GAUSSIAN_ELLIPSE"]
                        and name == "amp"
                    ):
                        num_param = len(kwargs_fixed["sigma"])
                        num += num_param
                        for i in range(num_param):
                            name_list.append(str(name + "_" + self._type + str(k)))
                    else:
                        num += 1
                        name_list.append(str(name + "_" + self._type + str(k)))
        return num, name_list

    def num_param_linear(self):
        """
        :return: number of linear basis set coefficients
        """
        return self._lightModel.num_param_linear(kwargs_list=self.kwargs_fixed)



__author__ = "sibirrer"

import numpy as np

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
class Slit(object):
    """Slit aperture description."""

    def __init__(self, length, width, center_ra=0, center_dec=0, angle=0):
        """

        :param length: length of slit
        :param width: width of slit
        :param center_ra: center of slit
        :param center_dec: center of slit
        :param angle: orientation angle of slit, angle=0 corresponds length in RA direction
        """
        self._length = length
        self._width = width
        self._center_ra, self._center_dec = center_ra, center_dec
        self._angle = angle

    def aperture_select(self, ra, dec):
        """

        :param ra: angular coordinate of photon/ray
        :param dec: angular coordinate of photon/ray
        :return: bool, True if photon/ray is within the slit, False otherwise
        """
        return (
            slit_select(
                ra,
                dec,
                self._length,
                self._width,
                self._center_ra,
                self._center_dec,
                self._angle,
            ),
            0,
        )

    @property
    def num_segments(self):
        """Number of segments with separate measurements of the velocity dispersion.

        :return: int
        """
        return 1


@export
def slit_select(ra, dec, length, width, center_ra=0, center_dec=0, angle=0):
    """

    :param ra: angular coordinate of photon/ray
    :param dec: angular coordinate of photon/ray
    :param length: length of slit
    :param width: width of slit
    :param center_ra: center of slit
    :param center_dec: center of slit
    :param angle: orientation angle of slit, angle=0 corresponds length in RA direction
    :return: bool, True if photon/ray is within the slit, False otherwise
    """
    ra_ = ra - center_ra
    dec_ = dec - center_dec
    x = np.cos(angle) * ra_ + np.sin(angle) * dec_
    y = -np.sin(angle) * ra_ + np.cos(angle) * dec_

    if abs(x) < length / 2.0 and abs(y) < width / 2.0:
        return True
    else:
        return False


@export
class Frame(object):
    """Rectangular box with a hole in the middle (also rectangular), effectively a
    frame."""

    def __init__(self, width_outer, width_inner, center_ra=0, center_dec=0, angle=0):
        """

        :param width_outer: width of box to the outer parts
        :param width_inner: width of inner removed box
        :param center_ra: center of slit
        :param center_dec: center of slit
        :param angle: orientation angle of slit, angle=0 corresponds length in RA direction
        """
        self._width_outer = width_outer
        self._width_inner = width_inner
        self._center_ra, self._center_dec = center_ra, center_dec
        self._angle = angle

    def aperture_select(self, ra, dec):
        """

        :param ra: angular coordinate of photon/ray
        :param dec: angular coordinate of photon/ray
        :return: bool, True if photon/ray is within the slit, False otherwise
        """
        return (
            frame_select(
                ra,
                dec,
                self._width_outer,
                self._width_inner,
                self._center_ra,
                self._center_dec,
                self._angle,
            ),
            0,
        )

    @property
    def num_segments(self):
        """Number of segments with separate measurements of the velocity dispersion.

        :return: int
        """
        return 1


@export
def frame_select(ra, dec, width_outer, width_inner, center_ra=0, center_dec=0, angle=0):
    """

    :param ra: angular coordinate of photon/ray
    :param dec: angular coordinate of photon/ray
    :param width_outer: width of box to the outer parts
    :param width_inner: width of inner removed box
    :param center_ra: center of slit
    :param center_dec: center of slit
    :param angle: orientation angle of slit, angle=0 corresponds length in RA direction
    :return: bool, True if photon/ray is within the box with a hole, False otherwise
    """
    ra_ = ra - center_ra
    dec_ = dec - center_dec
    x = np.cos(angle) * ra_ + np.sin(angle) * dec_
    y = -np.sin(angle) * ra_ + np.cos(angle) * dec_
    if abs(x) < width_outer / 2.0 and abs(y) < width_outer / 2.0:
        if abs(x) < width_inner / 2.0 and abs(y) < width_inner / 2.0:
            return False
        else:
            return True
    return False


@export
class Shell(object):
    """Shell aperture."""

    def __init__(self, r_in, r_out, center_ra=0, center_dec=0):
        """

        :param r_in: innermost radius to be selected
        :param r_out: outermost radius to be selected
        :param center_ra: center of the sphere
        :param center_dec: center of the sphere
        """
        self._r_in, self._r_out = r_in, r_out
        self._center_ra, self._center_dec = center_ra, center_dec

    def aperture_select(self, ra, dec):
        """

        :param ra: angular coordinate of photon/ray
        :param dec: angular coordinate of photon/ray
        :return: bool, True if photon/ray is within the slit, False otherwise
        """
        return (
            shell_select(
                ra, dec, self._r_in, self._r_out, self._center_ra, self._center_dec
            ),
            0,
        )

    @property
    def num_segments(self):
        """Number of segments with separate measurements of the velocity dispersion.

        :return: int
        """
        return 1


@export
def shell_select(ra, dec, r_in, r_out, center_ra=0, center_dec=0):
    """

    :param ra: angular coordinate of photon/ray
    :param dec: angular coordinate of photon/ray
    :param r_in: innermost radius to be selected
    :param r_out: outermost radius to be selected
    :param center_ra: center of the sphere
    :param center_dec: center of the sphere
    :return: boolean, True if within the radial range, False otherwise
    """
    x = ra - center_ra
    y = dec - center_dec
    r = np.sqrt(x**2 + y**2)
    if (r >= r_in) and (r < r_out):
        return True
    else:
        return False


@export
class IFUShells(object):
    """Class for an Integral Field Unit spectrograph with azimuthal shells where the
    kinematics are measured."""

    def __init__(self, r_bins, center_ra=0, center_dec=0):
        """

        :param r_bins: array of radial bins to average the dispersion spectra in ascending order.
         It starts with the innermost edge to the outermost edge.
        :param center_ra: center of the sphere
        :param center_dec: center of the sphere
        """
        self._r_bins = r_bins
        self._center_ra, self._center_dec = center_ra, center_dec

    def aperture_select(self, ra, dec):
        """

        :param ra: angular coordinate of photon/ray
        :param dec: angular coordinate of photon/ray
        :return: bool, True if photon/ray is within the slit, False otherwise, index of shell
        """
        return shell_ifu_select(
            ra, dec, self._r_bins, self._center_ra, self._center_dec
        )

    @property
    def num_segments(self):
        """Number of segments with separate measurements of the velocity dispersion
        :return: int."""
        return len(self._r_bins) - 1


@export
def shell_ifu_select(ra, dec, r_bin, center_ra=0, center_dec=0):
    """

    :param ra: angular coordinate of photon/ray
    :param dec: angular coordinate of photon/ray
    :param r_bin: array of radial bins to average the dispersion spectra in ascending order.
     It starts with the inner-most edge to the outermost edge.
    :param center_ra: center of the sphere
    :param center_dec: center of the sphere
    :return: boolean, True if within the radial range, False otherwise
    """
    x = ra - center_ra
    y = dec - center_dec
    r = np.sqrt(x**2 + y**2)
    for i in range(0, len(r_bin) - 1):
        if (r >= r_bin[i]) and (r < r_bin[i + 1]):
            return True, i
    return False, None



__author__ = "sibirrer"

import lenstronomy.Util.multi_gauss_expansion as mge

import numpy as np
import numpy.testing as npt
from lenstronomy.LightModel.Profiles.sersic import Sersic
from lenstronomy.LightModel.Profiles.hernquist import Hernquist
from lenstronomy.LightModel.Profiles.gaussian import MultiGaussian
import pytest


class TestMGE(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        self.sersic = Sersic()
        self.multiGaussian = MultiGaussian()

    def test_mge_1d_sersic(self):
        n_comp = 30
        r_sersic = 1.0
        n_sersic = 3.7
        I0_sersic = 1.0
        rs = np.logspace(-2.0, 1.0, 50) * r_sersic
        ss = self.sersic.function(
            rs, np.zeros_like(rs), amp=I0_sersic, n_sersic=n_sersic, R_sersic=r_sersic
        )

        amplitudes, sigmas, norm = mge.mge_1d(rs, ss, N=n_comp)
        ss_mge = self.multiGaussian.function(
            rs, np.zeros_like(rs), amp=amplitudes, sigma=sigmas
        )
        # print((ss - ss_mge)/ss)
        for i in range(10, len(ss) - 10):
            # print(rs[i])
            npt.assert_almost_equal((ss_mge[i] - ss[i]) / ss[i], 0, decimal=1)

        amplitudes, sigmas, norm = mge.mge_1d(rs, np.zeros_like(rs), N=n_comp)
        assert amplitudes[0] == 0

        amplitudes, sigmas, norm = mge.mge_1d(rs, np.zeros_like(rs), N=0)
        assert amplitudes[0] == 0

    def test_mge_sersic_radius(self):
        n_comp = 30
        r_sersic = 0.5
        n_sersic = 3.7
        I0_sersic = 1.0
        rs = np.logspace(-2.0, 1.0, 50) * r_sersic
        ss = self.sersic.function(
            rs, np.zeros_like(rs), amp=I0_sersic, n_sersic=n_sersic, R_sersic=r_sersic
        )

        amplitudes, sigmas, norm = mge.mge_1d(rs, ss, N=n_comp)
        ss_mge = self.multiGaussian.function(
            rs, np.zeros_like(rs), amp=amplitudes, sigma=sigmas
        )
        print((ss - ss_mge) / (ss + ss_mge))
        for i in range(10, len(ss) - 10):
            # print(rs[i])
            npt.assert_almost_equal((ss_mge[i] - ss[i]) / (ss[i]), 0, decimal=1)

    def test_mge_sersic_n_sersic(self):
        n_comp = 20
        r_sersic = 1.5
        n_sersic = 0.5
        I0_sersic = 1.0
        rs = np.logspace(-2.0, 1.0, 50) * r_sersic
        ss = self.sersic.function(
            rs, np.zeros_like(rs), amp=I0_sersic, n_sersic=n_sersic, R_sersic=r_sersic
        )

        amplitudes, sigmas, norm = mge.mge_1d(rs, ss, N=n_comp)
        ss_mge = self.multiGaussian.function(
            rs, np.zeros_like(rs), amp=amplitudes, sigma=sigmas
        )
        for i in range(10, len(ss) - 10):
            npt.assert_almost_equal(
                (ss_mge[i] - ss[i]) / (ss[i] + ss_mge[i]), 0, decimal=1
            )

        n_comp = 20
        r_sersic = 1.5
        n_sersic = 3.5
        I0_sersic = 1.0
        rs = np.logspace(-2.0, 1.0, 50) * r_sersic
        ss = self.sersic.function(
            rs, np.zeros_like(rs), amp=I0_sersic, n_sersic=n_sersic, R_sersic=r_sersic
        )

        amplitudes, sigmas, norm = mge.mge_1d(rs, ss, N=n_comp)
        ss_mge = self.multiGaussian.function(
            rs, np.zeros_like(rs), amp=amplitudes, sigma=sigmas
        )
        for i in range(10, len(ss) - 10):
            npt.assert_almost_equal(
                (ss_mge[i] - ss[i]) / (ss[i] + ss_mge[i]), 0, decimal=1
            )

    def test_hernquist(self):
        hernquist = Hernquist()
        n_comp = 20
        sigma0 = 1
        r_eff = 1.5
        rs = np.logspace(-2.0, 1.0, 50) * r_eff * 0.5
        ss = hernquist.function(rs, np.zeros_like(rs), sigma0, Rs=r_eff)
        amplitudes, sigmas, norm = mge.mge_1d(rs, ss, N=n_comp)
        ss_mge = self.multiGaussian.function(
            rs, np.zeros_like(rs), amp=amplitudes, sigma=sigmas
        )
        for i in range(10, len(ss) - 10):
            npt.assert_almost_equal(
                (ss_mge[i] - ss[i]) / (ss[i] + ss_mge[i]), 0, decimal=2
            )

    def test_hernquist_deprojection(self):
        hernquist = Hernquist()
        n_comp = 20
        sigma0 = 1
        r_eff = 1.5
        rs = np.logspace(-2.0, 1.0, 50) * r_eff * 0.5
        ss = hernquist.function(rs, np.zeros_like(rs), sigma0, Rs=r_eff)
        amplitudes, sigmas, norm = mge.mge_1d(rs, ss, N=n_comp)
        amplitudes_3d, sigmas_3d = mge.de_projection_3d(amplitudes, sigmas)
        ss_3d_mge = self.multiGaussian.function(
            rs, np.zeros_like(rs), amp=amplitudes_3d, sigma=sigmas_3d
        )
        ss_3d_mulit = self.multiGaussian.light_3d(rs, amp=amplitudes, sigma=sigmas)
        for i in range(10, len(ss_3d_mge)):
            npt.assert_almost_equal(
                (ss_3d_mge[i] - ss_3d_mulit[i]) / (ss_3d_mulit[i] + ss_3d_mge[i]),
                0,
                decimal=1,
            )

        ss_3d = hernquist.light_3d(rs, sigma0, Rs=r_eff)
        for i in range(10, len(ss_3d) - 10):
            npt.assert_almost_equal(
                (ss_3d_mge[i] - ss_3d[i]) / (ss_3d[i] + ss_3d_mge[i]), 0, decimal=1
            )

    def test_spemd(self):
        from lenstronomy.LensModel.Profiles.spep import SPEP
        from lenstronomy.LensModel.Profiles.multi_gaussian_kappa import (
            MultiGaussianKappa,
        )

        spep = SPEP()
        mge_kappa = MultiGaussianKappa()
        n_comp = 8
        theta_E = 1.41
        kwargs = {"theta_E": theta_E, "e1": 0, "e2": 0, "gamma": 1.61}
        rs = np.logspace(-2.0, 1.0, 100) * theta_E
        f_xx, f_xy, f_yx, f_yy = spep.hessian(rs, 0, **kwargs)
        kappa = 1 / 2.0 * (f_xx + f_yy)
        amplitudes, sigmas, norm = mge.mge_1d(rs, kappa, N=n_comp)
        kappa_mge = self.multiGaussian.function(
            rs, np.zeros_like(rs), amp=amplitudes, sigma=sigmas
        )
        f_xx_mge, f_xy_mge, f_yx_mge, f_yy_mge = mge_kappa.hessian(
            rs, np.zeros_like(rs), amp=amplitudes, sigma=sigmas
        )
        for i in range(0, 80):
            npt.assert_almost_equal(
                kappa_mge[i], 1.0 / 2 * (f_xx_mge[i] + f_yy_mge[i]), decimal=1
            )
            npt.assert_almost_equal((kappa[i] - kappa_mge[i]) / kappa[i], 0, decimal=1)

        f_ = spep.function(theta_E, 0, **kwargs)
        f_mge = mge_kappa.function(theta_E, 0, sigma=sigmas, amp=amplitudes)
        npt.assert_almost_equal(f_mge / f_, 1, decimal=2)

    def test_example(self):
        n_comp = 10
        rs = np.array(
            [
                0.01589126,
                0.01703967,
                0.01827108,
                0.01959148,
                0.0210073,
                0.02252544,
                0.02415329,
                0.02589879,
                0.02777042,
                0.02977731,
                0.03192923,
                0.03423667,
                0.03671086,
                0.03936385,
                0.04220857,
                0.04525886,
                0.0485296,
                0.0520367,
                0.05579724,
                0.05982956,
                0.06415327,
                0.06878945,
                0.07376067,
                0.07909115,
                0.08480685,
                0.09093561,
                0.09750727,
                0.10455385,
                0.11210966,
                0.12021152,
                0.12889887,
                0.13821403,
                0.14820238,
                0.15891255,
                0.17039672,
                0.18271082,
                0.19591482,
                0.21007304,
                0.22525444,
                0.24153295,
                0.25898787,
                0.2777042,
                0.29777311,
                0.31929235,
                0.34236672,
                0.36710861,
                0.39363853,
                0.42208569,
                0.45258865,
                0.48529597,
                0.52036697,
                0.55797244,
                0.59829556,
                0.64153272,
                0.6878945,
                0.73760673,
                0.79091152,
                0.8480685,
                0.90935605,
                0.97507269,
                1.04553848,
                1.12109664,
                1.20211518,
                1.28898871,
                1.38214034,
                1.48202378,
                1.58912553,
                1.70396721,
                1.82710819,
                1.95914822,
                2.10073042,
                2.25254437,
                2.4153295,
                2.58987865,
                2.77704199,
                2.9777311,
                3.19292345,
                3.42366716,
                3.67108607,
                3.93638527,
                4.22085689,
                4.5258865,
                4.85295974,
                5.20366966,
                5.57972441,
                5.98295559,
                6.41532717,
                6.87894505,
                7.37606729,
                7.90911519,
                8.48068497,
                9.09356051,
                9.75072687,
                10.45538481,
                11.21096643,
                12.02115183,
                12.88988708,
                13.82140341,
                14.82023784,
                15.89125526,
            ]
        )
        kappa = np.array(
            [
                12.13776067,
                11.60484966,
                11.09533396,
                10.60818686,
                10.14242668,
                9.69711473,
                9.27135349,
                8.86428482,
                8.47508818,
                8.10297905,
                7.7472073,
                7.40705574,
                7.08183863,
                6.77090034,
                6.47361399,
                6.18938022,
                5.917626,
                5.65780342,
                5.40938864,
                5.1718808,
                4.94480104,
                4.72769151,
                4.52011448,
                4.3216514,
                4.13190214,
                3.9504841,
                3.77703149,
                3.61119459,
                3.45263901,
                3.30104507,
                3.1561071,
                3.01753287,
                2.88504297,
                2.75837025,
                2.63725931,
                2.52146595,
                2.41075668,
                2.30490829,
                2.20370736,
                2.10694982,
                2.01444058,
                1.92599312,
                1.84142909,
                1.76057799,
                1.6832768,
                1.60936965,
                1.53870751,
                1.47114792,
                1.40655465,
                1.34479745,
                1.28575181,
                1.22929867,
                1.17532421,
                1.12371958,
                1.07438074,
                1.02720821,
                0.98210687,
                0.93898578,
                0.897758,
                0.85834039,
                0.82065349,
                0.78462129,
                0.75017114,
                0.71723359,
                0.68574222,
                0.65563353,
                0.62684681,
                0.59932403,
                0.57300967,
                0.5478507,
                0.52379638,
                0.5007982,
                0.47880979,
                0.45778683,
                0.43768691,
                0.41846951,
                0.40009589,
                0.38252899,
                0.3657334,
                0.34967525,
                0.33432216,
                0.31964317,
                0.30560868,
                0.29219041,
                0.27936129,
                0.26709545,
                0.25536817,
                0.24415579,
                0.23343571,
                0.22318631,
                0.21338694,
                0.20401782,
                0.19506006,
                0.18649562,
                0.17830721,
                0.17047832,
                0.16299318,
                0.15583668,
                0.14899441,
                0.14245255,
            ]
        )
        amplitudes, sigmas, norm = mge.mge_1d(rs, kappa, N=n_comp)

    def test_nfw_sersic(self):
        kwargs_lens_nfw = {"alpha_Rs": 1.4129647849966354, "Rs": 7.0991113634274736}
        kwargs_lens_sersic = {
            "k_eff": 0.24100561407593576,
            "n_sersic": 1.8058507329346063,
            "R_sersic": 1.0371803141813705,
        }
        from lenstronomy.LensModel.Profiles.nfw import NFW
        from lenstronomy.LensModel.Profiles.sersic import Sersic

        nfw = NFW()
        sersic = Sersic()
        theta_E = 1.5
        n_comp = 10
        rs = np.logspace(-2.0, 1.0, 100) * theta_E
        f_xx_nfw, f_xy_nfw, f_yx_nfw, f_yy_nfw = nfw.hessian(rs, 0, **kwargs_lens_nfw)
        f_xx_s, f_xy_s, f_yx_s, f_yy_s = sersic.hessian(rs, 0, **kwargs_lens_sersic)
        kappa = 1 / 2.0 * (f_xx_nfw + f_xx_s + f_yy_nfw + f_yy_s)
        amplitudes, sigmas, norm = mge.mge_1d(rs, kappa, N=n_comp)
        kappa_mge = self.multiGaussian.function(
            rs, np.zeros_like(rs), amp=amplitudes, sigma=sigmas
        )
        from lenstronomy.LensModel.Profiles.multi_gaussian_kappa import (
            MultiGaussianKappa,
        )

        mge_kappa = MultiGaussianKappa()
        f_xx_mge, f_xy_mge, f_yx_mge, f_yy_mge = mge_kappa.hessian(
            rs, np.zeros_like(rs), amp=amplitudes, sigma=sigmas
        )
        for i in range(0, 80):
            npt.assert_almost_equal(
                kappa_mge[i], 1.0 / 2 * (f_xx_mge[i] + f_yy_mge[i]), decimal=1
            )
            npt.assert_almost_equal((kappa[i] - kappa_mge[i]) / kappa[i], 0, decimal=1)

        f_nfw = nfw.function(theta_E, 0, **kwargs_lens_nfw)
        f_s = sersic.function(theta_E, 0, **kwargs_lens_sersic)
        f_mge = mge_kappa.function(theta_E, 0, sigma=sigmas, amp=amplitudes)
        npt.assert_almost_equal(f_mge / (f_nfw + f_s), 1, decimal=2)


if __name__ == "__main__":
    pytest.main()



import pytest
import numpy as np
import numpy.testing as npt
from lenstronomy.Sampling.Likelihoods.flux_ratio_likelihood import FluxRatioLikelihood
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.lens_model_extensions import LensModelExtensions
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver


class TestFluxRatioLikelihood(object):
    def setup_method(self):
        lens_model_list = ["SPEP", "SHEAR"]
        lensModel = LensModel(lens_model_list=lens_model_list)
        lensModelExtensions = LensModelExtensions(lensModel=lensModel)
        lensEquationSolver = LensEquationSolver(lensModel=lensModel)

        x_source, y_source = 0.02, 0.01
        kwargs_lens = [
            {
                "theta_E": 1.0,
                "e1": 0.1,
                "e2": 0.1,
                "gamma": 2.0,
                "center_x": 0,
                "center_y": 0,
            },
            {"gamma1": 0.06, "gamma2": -0.03},
        ]

        x_img, y_img = lensEquationSolver.image_position_from_source(
            kwargs_lens=kwargs_lens, sourcePos_x=x_source, sourcePos_y=y_source
        )
        print("image positions are: ", x_img, y_img)
        mag_inf = lensModel.magnification(x_img, y_img, kwargs_lens)
        print("point source magnification: ", mag_inf)

        source_size_arcsec = 0.001
        window_size = 0.1
        grid_number = 100
        print("source size in arcsec: ", source_size_arcsec)
        mag_finite = lensModelExtensions.magnification_finite(
            x_pos=x_img,
            y_pos=y_img,
            kwargs_lens=kwargs_lens,
            source_sigma=source_size_arcsec,
            window_size=window_size,
            grid_number=grid_number,
        )
        flux_ratios = mag_finite[1:] / mag_finite[0]
        flux_ratio_errors = [0.1, 0.1, 0.1]
        flux_ratio_cov = np.diag([0.1, 0.1, 0.1]) ** 2
        self.flux_likelihood = FluxRatioLikelihood(
            lens_model_class=lensModel,
            flux_ratios=flux_ratios,
            flux_ratio_errors=flux_ratio_errors,
            source_type="GAUSSIAN",
            window_size=window_size,
            grid_number=grid_number,
        )

        self.flux_likelihood_inf = FluxRatioLikelihood(
            lens_model_class=lensModel,
            flux_ratios=flux_ratios,
            flux_ratio_errors=flux_ratio_errors,
            source_type="INF",
            window_size=window_size,
            grid_number=grid_number,
        )
        self.flux_likelihood_inf_cov = FluxRatioLikelihood(
            lens_model_class=lensModel,
            flux_ratios=flux_ratios,
            flux_ratio_errors=flux_ratio_cov,
            source_type="INF",
            window_size=window_size,
            grid_number=grid_number,
        )
        self.kwargs_cosmo = {"source_size": source_size_arcsec}
        self.x_img, self.y_img = x_img, y_img
        self.kwargs_lens = kwargs_lens

    def test_logL(self):
        logL = self.flux_likelihood.logL(
            self.x_img, self.y_img, self.kwargs_lens, kwargs_special=self.kwargs_cosmo
        )
        assert logL == 0

        logL_inf = self.flux_likelihood_inf.logL(
            self.x_img, self.y_img, self.kwargs_lens, {}
        )
        npt.assert_almost_equal(logL_inf, 0, decimal=4)

    def test__logL(self):
        lensModel = LensModel(lens_model_list=[])
        flux_ratios_init = np.array([1.0, 1.0, 1.0])
        flux_ratio_errors = np.array([1.0, 1.0, 1.0])
        flux_likelihood = FluxRatioLikelihood(
            lens_model_class=lensModel,
            flux_ratios=flux_ratios_init,
            flux_ratio_errors=flux_ratio_errors,
        )

        flux_ratios = np.array([0, 1, np.nan])
        logL = flux_likelihood._logL(flux_ratios)
        assert logL == -(10**15)

        flux_likelihood = FluxRatioLikelihood(
            lens_model_class=lensModel,
            flux_ratios=flux_ratios_init,
            flux_ratio_errors=np.array([0.0, 1.0, 1.0]),
        )
        flux_ratios = np.array([1.0, 1.0, 1.0])
        logL = flux_likelihood._logL(flux_ratios)
        assert logL == -(10**15)

    def test_numimgs(self):
        # Test with a different number of images
        logL = self.flux_likelihood.logL(
            self.x_img[:-1],
            self.y_img[:-1],
            self.kwargs_lens,
            kwargs_special=self.kwargs_cosmo,
        )
        assert logL == -(10**15)

    def test_covmatrix(self):
        # Test with a different number of images
        logL = self.flux_likelihood_inf_cov.logL(
            self.x_img, self.y_img, self.kwargs_lens, kwargs_special=self.kwargs_cosmo
        )
        npt.assert_almost_equal(logL, 0, decimal=8)


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

from lenstronomy.GalKin.galkin_model import GalkinModel

import numpy.testing as npt
import pytest


class TestGalkinModel(object):
    def setup_method(self):
        pass

    def test_radius_slope_anisotropy(self):
        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}
        kwargs_model = {
            "anisotropy_model": "OM",
            "mass_profile_list": ["SPP"],
            "light_profile_list": ["HERNQUIST"],
        }
        kwargs_numerics = {
            "interpol_grid_num": 1000,
            "log_integration": True,
            "max_integrate": 100,
            "min_integrate": 0.001,
        }
        kin_analytic = GalkinModel(
            kwargs_model,
            kwargs_cosmo,
            analytic_kinematics=True,
            kwargs_numerics=kwargs_numerics,
        )
        r = 1
        theta_E, gamma = 1, 2.0
        a_ani = 10
        r_eff = 0.1
        out = kin_analytic.check_df(
            r,
            kwargs_mass={"theta_E": theta_E, "gamma": gamma},
            kwargs_light={"r_eff": r_eff},
            kwargs_anisotropy={"r_ani": a_ani * r_eff},
        )
        assert out > 0
        print(out)

        kin_numeric = GalkinModel(
            kwargs_model,
            kwargs_cosmo,
            analytic_kinematics=False,
            kwargs_numerics=kwargs_numerics,
        )
        out_num = kin_numeric.check_df(
            r,
            kwargs_mass=[{"theta_E": theta_E, "gamma": gamma}],
            kwargs_light=[{"Rs": r_eff * 0.551, "amp": 1}],
            kwargs_anisotropy={"r_ani": a_ani * r_eff},
        )
        assert out_num > 1
        npt.assert_almost_equal(out_num / out, 1, decimal=2)

        kin_numeric_default = GalkinModel(
            kwargs_model, kwargs_cosmo, analytic_kinematics=False, kwargs_numerics=None
        )


if __name__ == "__main__":
    pytest.main()



import numpy as np
import copy
import lenstronomy.Util.mask_util as mask_util
from lenstronomy.Util import param_util

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
def half_light_radius(lens_light, x_grid, y_grid, center_x=0, center_y=0):
    """

    :param lens_light: array of surface brightness
    :param x_grid: x-axis coordinates
    :param y_grid: y-axis coordinates
    :param center_x: center of light
    :param center_y: center of light
    :return:
    """
    lens_light[lens_light < 0] = 0
    total_flux_2 = np.sum(lens_light) / 2.0
    r_max = np.max(np.sqrt((x_grid - center_x) ** 2 + (y_grid - center_y) ** 2))
    for i in range(1000):
        r = i / 500.0 * r_max
        mask = mask_util.mask_azimuthal(x_grid, y_grid, center_x, center_y, r)
        flux_enclosed = np.sum(np.array(lens_light) * mask)
        if flux_enclosed > total_flux_2:
            return r
    return -1


@export
def radial_profile(light_grid, x_grid, y_grid, center_x=0, center_y=0, n=None):
    """Computes radial profile.

    :param light_grid: array of surface brightness
    :param x_grid: x-axis coordinates
    :param y_grid: y-axis coordinates
    :param center_x: center of light
    :param center_y: center of light
    :param n: number of discrete steps
    :return: I(r), r with r in units of the coordinate grid
    """
    r_max = np.max(np.sqrt((x_grid - center_x) ** 2 + (y_grid - center_y) ** 2))
    if n is None:
        n = int(np.sqrt(len(x_grid)))
    I_r = np.zeros(n)
    I_enclosed = 0
    r = np.linspace(1.0 / n * r_max, r_max, n)
    for i, r_i in enumerate(r):
        mask = mask_util.mask_azimuthal(x_grid, y_grid, center_x, center_y, r_i)
        flux_enclosed = np.sum(np.array(light_grid) * mask)
        I_r[i] = flux_enclosed - I_enclosed
        I_enclosed = flux_enclosed
    return I_r, r


@export
def azimuthalAverage(image, center=None):
    """Calculate the azimuthally averaged radial profile.

    image - The 2D image
    center - The [x,y] pixel coordinates used as the center. The default is None, which then uses the center of the
    image (including fractional pixels).
    :return: I(r) (averaged), r of bin edges in units of pixels of the 2D image
    """
    # Calculate the indices from the image
    y, x = np.indices(image.shape)

    if not center:
        center = np.array([(x.max() - x.min()) / 2.0, (x.max() - x.min()) / 2.0])

    r = np.hypot(x - center[0], y - center[1])

    # Get sorted radii
    ind = np.argsort(r.flat)
    r_sorted = r.flat[ind]
    i_sorted = image.flat[ind]

    # Get the integer part of the radii (bin size = 1)
    r_int = r_sorted.astype(int)

    # Find all pixels that fall within each radial bin.
    deltar = r_int[1:] - r_int[:-1]  # Assumes all radii represented
    rind = np.where(deltar)[0]  # location of changed radius
    nr = rind[1:] - rind[:-1]  # number of radius bin

    # Cumulative sum to figure out sums for each radius bin
    csim = np.cumsum(i_sorted, dtype=float)
    tbin = csim[rind[1:]] - csim[rind[:-1]]
    r_bin = np.linspace(start=1, stop=len(tbin) + 1 - 0.5, num=len(tbin))
    radial_prof = tbin / nr
    return radial_prof, r_bin


@export
def moments(I_xy_input, x, y):
    """Compute quadrupole moments from a light distribution.

    :param I_xy_input: light distribution
    :param x: x-coordinates of I_xy
    :param y: y-coordinates of I_xy
    :return: Q_xx, Q_xy, Q_yy
    """
    I_xy = copy.deepcopy(I_xy_input)
    background = np.minimum(0, np.min(I_xy))
    I_xy -= background
    x_ = np.sum(I_xy * x)
    y_ = np.sum(I_xy * y)
    r = (np.max(x) - np.min(x)) / 3.0
    mask = mask_util.mask_azimuthal(x, y, center_x=x_, center_y=y_, r=r)
    Q_xx = np.sum(I_xy * mask * (x - x_) ** 2)
    Q_xy = np.sum(I_xy * mask * (x - x_) * (y - y_))
    Q_yy = np.sum(I_xy * mask * (y - y_) ** 2)
    return Q_xx, Q_xy, Q_yy, background / np.mean(I_xy)


@export
def _ellipticities(I_xy, x, y):
    """Compute ellipticities of a light distribution.

    :param I_xy: surface brightness I(x, y) as array
    :param x: x-coordinates in same shape as I_xy
    :param y: y-coordinates in same shape as I_xy
    :return: reduced shear moments g1, g2
    """
    Q_xx, Q_xy, Q_yy, bkg = moments(I_xy, x, y)
    norm = Q_xx + Q_yy + 2 * np.sqrt(Q_xx * Q_yy - Q_xy**2)
    e1 = (Q_xx - Q_yy) / norm
    e2 = 2 * Q_xy / norm
    return e1 / (1 + bkg), e2 / (1 + bkg)


@export
def ellipticities(
    I_xy, x_grid, y_grid, num_iterative=30, iterative=False, center_x=0, center_y=0
):
    """

    :param I_xy: surface brightness I(x, y) as array
    :param x_grid: x-coordinates in same shape as I_xy
    :param y_grid: y-coordinates in same shape as I_xy
    :param iterative: if True iteratively adopts an eccentric mask to overcome edge effects
    :type iterative: boolean
    :param num_iterative: number of iterative changes in ellipticity
    :type num_iterative: int
    :return: e1, e2 eccentricities
    """
    radius = (np.max(x_grid) - np.min(x_grid)) / 2.0
    mask = mask_util.mask_azimuthal(
        x_grid, y_grid, center_x=center_x, center_y=center_y, r=radius
    )
    e1, e2 = _ellipticities(I_xy * mask, x_grid - center_x, y_grid - center_y)
    phi, q = param_util.ellipticity2phi_q(e1, e2)
    if iterative:
        for i in range(num_iterative):
            mask = mask_util.mask_eccentric(
                x_grid, y_grid, center_x, center_y, e1, e2, r=radius * q / np.sqrt(2)
            )
            e1, e2 = _ellipticities(I_xy * mask, x_grid - center_x, y_grid - center_y)
            phi, q = param_util.ellipticity2phi_q(e1, e2)
    return e1, e2


@export
def bic_model(logL, num_data, num_param):
    """Bayesian information criteria.

    :param logL: log likelihood value
    :param num_data: numbers of data
    :param num_param: numbers of model parameters
    :return: BIC value
    """
    bic = -2 * logL + (np.log(num_data) * num_param)
    return bic


@export
def profile_center(kwargs_list, center_x=None, center_y=None):
    """Utility routine that results in the centroid estimate for the profile estimates.

    :param kwargs_list: light parameter keyword argument list (can be light or mass)
    :param center_x: None or center
    :param center_y: None or center
    :return: center_x, center_y
    """
    if center_x is None or center_y is None:
        if "center_x" in kwargs_list[0]:
            center_x = kwargs_list[0]["center_x"]
            center_y = kwargs_list[0]["center_y"]
        else:
            raise ValueError(
                "The center has to be provided as a function argument or the first profile in the list"
                " must come with a center."
            )
    return center_x, center_y
